import unittest
import numpy as np
from radbm.utils.torch import torch_cast_cpu

class TorchCast(unittest.TestCase):
    def test_torch_cast_cpu(self):
        a = np.zeros((4,5), dtype=np.float32)
        b = torch_cast_cpu(a)
        self.assertTrue((a==b.numpy()).all())



import unittest, torch
import numpy as np
from radbm.search.radius import HammingRadiusSDS

class TestHammingRadiusSDS(unittest.TestCase):
    def test_numpy_cast(self):
        mbht = HammingRadiusSDS(64, 2)
        mbht.batch_insert(torch.zeros((10,64),dtype=torch.bool), range(10))
        
    def test_hammingradius(self):
        sds = HammingRadiusSDS(4, 2)
        #(0 in 4) + (1 in 4) + (2 in 4) = 1 + 4 + 6 = 11
        self.assertEqual(sds.masks.shape, (11, 4))
        
        documents = np.array([
            [0,0,0,0], #0
            [0,0,1,1], #1
            [0,1,0,1], #2
            [1,1,1,0], #3
        ], dtype=np.bool)
        
        queries = np.array([
            [0,0,0,0], # match with 0, 1, 2
            [1,0,0,0], # match with 0, 3
            [0,1,1,0], # match with 0, 1, 2, 3
        ], dtype=np.bool)
        
        #insert + search
        sds.batch_insert(documents, range(4))
        indexes = sds.batch_search(queries)
        expected = [{0,1,2}, {0,3}, {0,1,2,3}]
        self.assertEqual(indexes, expected)
        
        #itersearch with yield_empty=True
        indexes = list(sds.itersearch(queries[0], yield_empty=True))
        expected = [
            {0}, #radius 0
            set(), set(), set(), set(), #radius 1
            set(), set(), set(), set(), {2}, {1}, #radius 2
        ]
        self.assertEqual(indexes, expected)
        
        #itersearch with yield_empty=False
        indexes = list(sds.itersearch(queries[0], yield_empty=False))
        expected = [{0}, {2}, {1}] #same order as above
        self.assertEqual(indexes, expected)


import torch
from radbm.search.base import BaseSDS
logsigmoid = torch.nn.LogSigmoid()

class EfficientLearnableBinaryAccess(BaseSDS, torch.nn.Module):
    """
    EfficientLearnableBinaryAccess (ELBA) is a base class for concrete models.
    Given a search data structure and two parametric encoding functions, one
    for the queries (fq) and one for the documents (fd), both producing a 
    Multi-Bernoulli code (i.e. in [0,1]^n) in its logits form (pre sigmoid).
    ELBM uses the structure to store and retrieve data using the Multi-Bernoulli
    code.
    
    Parameters
    ----------
    fq : torch.nn.Module
        The parametric function of the queries outputting in logits (pre sigmoid).
    fd : torch.nn.Module
        The parametric function of the documents outputting in logits (pre sigmoid).
    struct : BaseSDS subclass
        The structure used for storing and retrieval.
    optim : torch.optim (optional)
        The optimizer (minimizer) class used for the parametric function.
        (default torch.optim.Adam)
    lr : float (optional)
        The learning rate of the optimizer. (default 0.001)
    """
    def __init__(self, fq, fd, struct, optim=torch.optim.Adam, lr=0.001):
        torch.nn.Module.__init__(self)
        self.fq = fq
        self.fd = fd
        self.struct = struct
        self.optim = optim(self.parameters(), lr)
        
    def _log_sigmoid_pairs(self, logits):
        return torch.stack([logsigmoid(-logits), logsigmoid(logits)], dim=1)
        
    def batch_insert(self, documents, indexes, *args, **kwargs):
        """
        Insert the index of each documents in the data structure
        
        Parameters
        ----------
        documents : torch.tensor
            The documents to insert the first dim being the batch.
        indexes : iterable of hashable
            most notable example is a list of int. len(indexes) most
            be equal to len(documents).
        *args
            passed to self.struct.batch_insert
        **kwargs
            passed to self.struct.batch_insert
            
        Returns
        -------
        self
        """
        dmb = self._log_sigmoid_pairs(self.fd(documents))
        self.struct.batch_insert(dmb, indexes, *args, **kwargs)
        return self
        
    def batch_search(self, queries, *args, **kwargs):
        """
        Search in the data structure for the relevant indexes for each queries.
        
        Parameters
        ----------
        queries : torch.tensor
            The search queries, the first dim being the batch.
        *args
            passed to self.struct.batch_search
        **kwargs
            passed to self.struct.batch_search
            
        Returns
        -------
        indexes_list : list of (set or list)
            Is the list of the relevant indexes for each queries. 
            len(indexes_list) = len(queries).
        """
        qmb = self._log_sigmoid_pairs(self.fq(queries))
        return self.struct.batch_search(qmb, *args, **kwargs)
        
    def batch_itersearch(self, queries, *args, **kwargs):
        """
        Iteratively search in the data structure for the relevant
        indexes for each queries.
        
        Parameters
        ----------
        queries : torch.tensor
            The search queries, the first dim being the batch.
        *args
            passed to self.struct.batch_itersearch
        **kwargs
            passed to self.struct.batch_itersearch
            
        Returns
        -------
        generator_list : list of generators (of set or list)
            Each generator yield relevant indexes for the corresponding queries. 
            len(generator_list) = len(queries).
        """
        qmb = self._log_sigmoid_pairs(self.fq(queries))
        return self.struct.batch_itersearch(qmb, *args, **kwargs)
    
    def get_state(self):
        return {
            'f': self.state_dict(),
            'optim': self.optim.state_dict(),
            'struct': self.struct.get_state(),
        }
    
    def set_state(self, state):
        self.load_state_dict(state['f'])
        self.optim.load_state_dict(state['optim'])
        self.struct.set_state(state['struct'])
        return self


import h5py, torch
import numpy as np
from radbm.loaders.base import IRLoader
from radbm.utils.fetch import fetch_file
from .alterimages import AlterImages

def coco_finder(which, path=None):
    file_paths = fetch_file(
        which,
        path,
        data_type='dataset',
        subdirs=['MSCoco', 'Coco'],
        download=False
    )
    if not file_paths:
        raise FileNotFoundError('could not locate squared_coco.hdf5')
    return h5py.File(file_paths[0], 'r')

class AlteredCoco(IRLoader):
    def __init__(
        self, batch_global_scale_range=[1.,1.], generator=torch.default_generator, path=None,
        mode='relational_matrix', which='train', backend='numpy', device='cpu', rng=np.random):
        super().__init__(mode=mode, which=which, backend=backend, device=device, rng=rng)
        if mode != 'relational_matrix':
            raise NotImplementedError('mode={} not implemented yet'.format(mode))
        self.original_coco = coco_finder('squared_coco.hdf5', path)
        self.altered_coco = coco_finder('altered_coco.hdf5', path)
        self.alter = AlterImages((256, 256), batch_global_scale_range=batch_global_scale_range, generator=generator)
        N = len(self.original_coco['data'])
        self.train_group = {
            'original_indexes':np.arange(60000,N-40000),
            'altered_indexes': np.arange(0,10000)
        }
        self.valid_group = {
            'original_indexes':np.array(list(range(0,60000)) + list(range(N-40000,N))),
            'altered_indexes': np.arange(0,10000)
        }
        getattr(self, self.which)()
    
    def close(self):
        self.original_coco.close()
        self.altered_coco.close()
        
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False
    
    def get_available_modes(self):
        return {
            #'relational_pairs', #(x,k,r)
            #'relational_triplets', #(x, pos_x, neg_x)
            'relational_matrix', #(xs, ks, R)
        }
    
    def get_relation_prob(self):
        return 1/100000
    
    def get_relation_log_prob(self):
        return -np.log(100000)
    
    def iter_queries(self, batch_size, maximum=np.inf):
        which = self.which
        coco_id = '{}_coco_id'.format(which)
        N = min(maximum, len(self.altered_indexes))
        nbatch = int(np.ceil(N//batch_size))
        for start in [i*batch_size for i in range(nbatch)]:
            if self.which != which:
                msg = 'self.which changed while iterating, should be "{}" but is now "{}".'
                raise RuntimeError(msg.format(which, self.which))
            end = min(start+batch_size, N)
            indexes = list(self.altered_indexes[start:end])
            data = self.altered_coco[which][indexes]/255
            relevant = self.altered_coco[coco_id][indexes]
            yield self.dynamic_cast(data), [{int(r)} for r in relevant]
    
    def iter_documents(self, batch_size, maximum=np.inf):
        which = self.which
        N = min(maximum, len(self.original_indexes))
        nbatch = int(np.ceil(N//batch_size))
        for start in [i*batch_size for i in range(nbatch)]:
            if self.which != which:
                msg = 'self.which changed while iterating, should be "{}" but is now "{}".'
                raise RuntimeError(msg.format(which, self.which))
            end = min(start+batch_size, N)
            indexes = list(self.original_indexes[start:end])
            data = self.original_coco['data'][indexes]/255
            ids = self.original_coco['coco_id'][indexes]
            yield self.dynamic_cast(data), list(ids)
            
    def choice_without_replacement(self, N, size):
        #faster than np.choice for size << N
        for _ in range(5): #not using while True because it is better late than never 
            ids = self.rng.rng.randint(0, N, size)
            if len(ids)==len(set(ids)):
                return ids
        return self.rng.rng.choice(range(N), size=size, replace=False)
    
    def batch(self, size=None, index=None, replace=True):
        if index is not None or not replace:
            raise ValueError('not implemented yet')
        ids = self.choice_without_replacement(len(self.original_indexes), size)
        indexes = sorted(self.original_indexes[ids])
        d = self.original_coco['data'][indexes]/255
        d = torch.tensor(d, dtype=torch.float32, device=self.alter.device)
        q = self.alter(d)
        r = self.dynamic_cast(np.eye(size, dtype=np.bool))
        return self.dynamic_cast(q), self.dynamic_cast(d), r


import torch
import numpy as np
from radbm.utils.torch import torch_logsumexp

def _log_poisson_binomial(log_q0, log_q1):
    bs, n = log_q0.shape
    dtype, device = log_q0.dtype, log_q0.device
    ninf = torch.tensor(-np.inf, dtype=dtype, device=device).expand(1, bs)
    z = torch.zeros((1, bs), dtype=dtype, device=device)
    for n,(a,b) in enumerate(zip(log_q0.permute(1,0), log_q1.permute(1,0))):
        w0 = torch.cat([a + z, ninf], dim=0)
        w1 = torch.cat([ninf, b + z], dim=0)
        z = torch_logsumexp(w0, w1)
    return z.permute(1,0)

def log_poisson_binomial(log_q0, log_q1):
    """
    Computes the events log probabilities w.r.t. a batch of Poisson Binomial R.V.
    The computation is numerically stable.
    
    Parameters
    ----------
    log_q0 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be zero. The Poisson Binomial is considered to
        be on the last dim. shape=(a1,a2,a3,...,am,n) where n is the number of bits for each 
        Poisson Binomial. a1,a2,a3,...,am are arbitrary but should match with log_q1.
    
    log_q1 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be one. The Poisson Binomial is considered to
        be on the last dim. shape=(a1,a2,a3,...,am,n) where n is the number of bits for each 
        Poisson Binomial. a1,a2,a3,...,am are arbitrary but should match with log_q0.
        
    Returns
    -------
    log_pb : torch.tensor(dtype=torch.float)
        The log probability of each Poisson Binomial events. shape=(a1,a2,a3,...,am,n+1).
        log_pb[i1,i2,i3,...,am,k] is the log probability that the sum of the n Bernoulli
        with parameters log_q1.exp()[i1,i2,i3,...,am] gives k.
        
    Notes
    -----
    We should have (1-log_q0.exp()).log() = log_q1 in theory, hence the input of this function
    is over specified. But for numerical stability we need both log_q0 and log_q1 and it is not
    possible to compute log_q0 from log_q1 with numerical stability (and vice versa). This is why
    they are both required as input. Since in some cases, it is possible to compute log_q0 and
    log_q1 with numerical stability, i.e. log_q0 = log_sigmoid(-logits) and log_q1 = log_sigmoid(logits)
        
    """
    shape = log_q0.shape
    if shape != log_q1.shape:
        msg = 'log_q0.shape and log_q1.shape must be the same, got {} and {} respectively'
        raise ValueError(msg.format(shape, log_q1.shape))
    n = shape[-1]
    log_pb = _log_poisson_binomial(log_q0.view(-1,n), log_q1.view(-1,n))
    return log_pb.view(*shape[:-1], n+1)

def log_hamming_binomial(log_p10, log_p11, log_p20, log_p21):
    """
    Computes the log probabilities of each Hamming Binomial events, parameterized
    with p1 and p2.
    
    Parameters
    ----------
    log_p10 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be zero of the first random vector.
        The Hamming Binomial is considered to be on the last dim. shape=(a1,a2,a3,...,am,n)
        where n is the number of bits for each vectors. a1,a2,a3,...,am are arbitrary but
        should be broadcastable with the other inputs.
    
    log_p11 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be one of the first random vector.
        The Hamming Binomial is considered to be on the last dim. shape=(a1,a2,a3,...,am,n)
        where n is the number of bits for each vectors. a1,a2,a3,...,am are arbitrary but
        should be broadcastable with the other inputs.
        
    log_p20 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be zero of the second random vector.
        The Hamming Binomial is considered to be on the last dim. shape=(a1,a2,a3,...,am,n)
        where n is the number of bits for each vectors. a1,a2,a3,...,am are arbitrary but
        should be broadcastable with the other inputs.
    
    log_p21 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be one of the second random vector.
        The Hamming Binomial is considered to be on the last dim. shape=(a1,a2,a3,...,am,n)
        where n is the number of bits for each vectors. a1,a2,a3,...,am are arbitrary but
        should be broadcastable with the other inputs.
        
    Returns
    -------
    log_hb : torch.tensor (dtype=torch.float)
        The log probability of each Hamming Binomial events. shape=(a1,a2,a3,...,am,n+1).
        log_pb[i1,i2,i3,...,am,k] is the log probability that the hamming distance between
        the two random Multi-Bernoulli parameterized by log_p11.exp()[i1,i2,i3,...,am] and
        log_p21.exp()[i1,i2,i3,...,am] be k.
        
    Notes
    -----
    see log_poisson_binomial's notes.
    """
    log_q0 = torch_logsumexp(log_p11 + log_p21, log_p10 + log_p20)
    log_q1 = torch_logsumexp(log_p11 + log_p20, log_p10 + log_p21)
    return log_poisson_binomial(log_q0, log_q1)


import unittest
import numpy as np
from radbm.metrics.sswr import HaltingChronoSSWR, HaltingCounterSSWR, ChronoSSWR, CounterSSWR

def moc(N, K, k):
    #short for matching oracle cost
    return k*(N+1)/(K+1)

class TestSSWR(unittest.TestCase):
    def test_countersswr(self):
        #test 1
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1}, {2,3}, {4,5}])
        out = CounterSSWR(relevant, delta_gen, N=20)
        
        #            find 4 in {4,5} + |{0,1,2,3}|   +     #generator calls
        expected_out = (moc(2,1,1) + len({0,1,2,3}) + len([{0,1}, {2,3}, {4,5}])) / moc(20,5,5)
        self.assertEqual(out, expected_out)
        
        #test 2 (recall=0.8) (we need 4 out of 5)
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1}, {2,3}, {4,5}])
        out = CounterSSWR(relevant, delta_gen, N=20, recall=0.8)
        
        #            find 2,3 in {2,3} + |{0,1}| + #generator calls
        expected_out = (moc(2,2,2) + len({0,1}) + len([{0,1}, {2,3}])) / moc(20,5,4)
        self.assertEqual(out, expected_out)
        
        #test 3 (with bad candidates)
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1,10,11}, {2,3,12}, {4,5,13}])
        out = CounterSSWR(relevant, delta_gen, N=20)
        
        #            find 4 in {4,5,13} + |{0,1,10,11,2,3,12}| + #generator calls
        expected_out = (moc(3,1,1) +   len({0,1,10,11,2,3,12}) + 3) / moc(20,5,5)
        self.assertEqual(out, expected_out)
        
        #test 4 (with bad candidates + duplicates (and on_duplicate_candidates='ignore'))
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1,10,11}, {2,3,12,0,10}, {4,5,13,1,2,12}])
        out = CounterSSWR(relevant, delta_gen, N=20, on_duplicate_candidates='ignore')
        
        #            find 4 in {4,5,13} + |{0,1,10,11,2,3,12}| + #generator calls
        expected_out = (moc(3,1,1) +   len({0,1,10,11,2,3,12}) + 3) / moc(20,5,5)
        self.assertEqual(out, expected_out)
        
        #test 5 (halting + bad candidates + recall=5/6)
        relevant = {0,1,2,3,4,5}
        delta_gen = (s for s in [{0,1}, {3,10}, {12}])
        out = CounterSSWR(relevant, delta_gen, N=20, recall=5/6, allow_halt=True)
        
        #        find 2 out of 3 in 15 elements + |{0,1,10,11,2,3,12}| + #generator calls
        expected_out = (moc(15,3,2) + len({0,1,3,10,12}) + 3) / moc(20,6,5)
        self.assertEqual(out, (expected_out, True))
        
    def test_haltingcountersswr(self):
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1,5}, {2,3}, {4,6,7}])
        sswrs, halts = HaltingCounterSSWR(relevant, delta_gen, N=20, max_halt=4)
        expected_sswrs = np.array([
            moc(20,5,5) + 0 + 0, #halt
            moc(17,3,3) + 3 + 1, #halt
            moc(15,1,1) + 5 + 2, #halt
            moc(3,1,1) + 5 + 3, #not halt
            moc(3,1,1) + 5 + 3, #not halt (same->copy)
        ])/moc(20,5,5)
        expected_halts = np.array([1,1,1,0,0])
        self.assertTrue(np.allclose(sswrs, expected_sswrs))
        self.assertTrue(np.allclose(halts, expected_halts))
        
        #with recall = 6/7 and max_hatl=4
        relevant = {0,1,2,3,4,5,6}
        delta_gen = (s for s in [{0,1,7}, {2,3}, {4,8,9}, {10,11}, {5}]) #<-- 5 found at step 5<4
        sswrs, halts = HaltingCounterSSWR(relevant, delta_gen, N=20, max_halt=4, recall=6/7)
        expected_sswrs = np.array([
            moc(20,7,6) + 0 + 0, #halt
            moc(17,5,4) + 3 + 1, #halt
            moc(15,3,2) + 5 + 2, #halt
            moc(12,2,1) + 8 + 3, #halt
            moc(10,2,1) + 10 + 4, #halt
        ])/moc(20,7,6)
        expected_halts = np.array([1,1,1,1,1])
        self.assertTrue(np.allclose(sswrs, expected_sswrs))
        self.assertTrue(np.allclose(halts, expected_halts))
    
    def test_chronosswr(self):
        #make sure it runs
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1}, {2,3}, {4,5}])
        out = ChronoSSWR(relevant, delta_gen, N=20)
        
    def test_haltingchronosswr(self):
        #make sure it runs
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1}, {2,3}, {4,5}])
        out = HaltingChronoSSWR(relevant, delta_gen, N=20, max_halt=4)
        
    def test_generator_halt_error(self):
        relevant = {0,1,2,3,4}
        
        #test 1 (non-empty)
        delta_gen = (s for s in [{0,1}])
        with self.assertRaises(LookupError):
            CounterSSWR(relevant, delta_gen, N=20, on_duplicate_candidates='raise')
            
        #test 2 (empty)
        delta_gen = (s for s in [])
        with self.assertRaises(LookupError):
            CounterSSWR(relevant, delta_gen, N=20, on_duplicate_candidates='raise')
    
    def test_duplicate_candidates(self):
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1,10,11}, {2,3,12,0,10}, {4,5,13,1,2,12}])
        with self.assertRaises(RuntimeError):
            CounterSSWR(relevant, delta_gen, N=20, on_duplicate_candidates='raise')
    
    def test_bad_duplicate_candidates_option(self):
        relevant = {0,1,2,3,4}
        delta_gen = (s for s in [{0,1,10,11}, {2,3,12,0,10}, {4,5,13,1,2,12}])
        with self.assertRaises(ValueError):
            CounterSSWR(relevant, delta_gen, N=20, on_duplicate_candidates='bad_option')


import os, torch
import numpy as np
from radbm.utils.os import StateObj
from radbm.utils.torch import (
    TorchNumpy,
    TorchNumpyRNG,
    torch_cast_cpu
)

class Loader(StateObj):
    """
    An abstract class managing numpy vs torch, cpu vs gpu and train
    vs valid vs test. This should be subclassed with a particular
    dataset (i.e. Mnist).
    
    Parameters
    ----------
    which : str
        Which datasets version to use. Should be 'train', 'valid' or 'test'.
    backend : str
        Which backend to use, should be 'numpy' or 'torch'.
    device : str
        Which device to use, should be 'cpu' or 'cuda'. 'cuda' is only
        available if backend=='torch'.
    rng : numpy.random.RandomState
        A random number generator for reproducibility.
    """
    def __init__(self, which, backend, device, rng=None):
        rng = np.random if rng is None else rng
        self.value_check('which', {'train', 'valid', 'test'}, which)
        self.value_check('backend', {'numpy', 'torch'}, backend)
        self.value_check('device', {'cpu', 'cuda'}, device)
        
        if device=='cuda': self.cuda_check()
        if backend=='numpy' and device=='cuda':
            raise ValueError('cannot use cuda with numpy backend')
        
        self.rng=TorchNumpyRNG(rng=rng)
        self.switch_attrs = ['rng']
        self.which=which
        self.backend='numpy'
        self.device='cpu'
        if backend=='torch': self.torch()
        if device=='cuda': self.cuda()
            
        self.train_group = dict()
        self.valid_group = dict()
        self.test_group = dict()
        self.end_init()
        
    def end_init(self):
        getattr(self, self.which)()
    
    def value_check(self, name, valids, value):
        if value not in valids:
            msg = '{} must be in {}, got {}'
            raise ValueError(msg.format(name, valids, value))
            
    def cuda_check(self):
        if not torch.cuda.is_available():
            raise ValueError('cannot use cuda, it is not available')
    
    def register_switch(self, name, data):
        """
        This function should only be used when subclassing. This is to
        register data for when a user will call: numpy(), torch(), cpu()
        or cuda(). Each value will be transfered to the appropriate format.
        
        Parameters
        ----------
        name : str
            The name of the data. setarrt is used so one could later do self.<name>
            to reach the data.
        data : numpy.ndarray or torch.Tensor
            The data to register.
        """
        if hasattr(self, name):
            raise ValueError('{} already used'.format(name))
        tnp = TorchNumpy(data, self.backend, self.device)
        setattr(self, name, tnp)
        self.switch_attrs.append(name)
    
    def apply_switch(self, function_name):
        for name in self.switch_attrs:
            getattr(getattr(self, name), function_name)()
    
    def torch(self):
        """
        Converts each registered data (using register_switch) into torch.Tensor
        """
        if self.backend == 'torch': return self
        self.apply_switch('torch')
        self.backend = 'torch'
        return self
        
    def numpy(self):
        """
        Converts each registered data (using register_switch) into numpy.ndarray
        
        Raises
        ------
        ValueError
            If device=='cuda'
        """
        if self.backend == 'numpy': return self
        if self.device == 'cuda':
            msg = 'cannot use numpy backend with cuda, try loader.cpu() first'
            raise ValueError(msg)
        self.apply_switch('numpy')
        self.backend = 'numpy'
        return self
        
    def cuda(self):
        """
        Transfers each registered data (using register_switch) to the GPU
        
        Raises
        ------
        ValueError
            If backend=='numpy'
        """
        if self.device == 'cuda': return self
        if self.backend == 'numpy':
            msg = 'cannot use cuda with numpy backend, try loader.torch() first'
            raise ValueError(msg)
        self.cuda_check()
        self.apply_switch('cuda')
        self.device = 'cuda'
        return self
        
    def cpu(self):
        """
        Transfers each registered data (using register_switch) to the CPU
        """
        if self.device == 'cpu': return self
        self.apply_switch('cpu')
        self.device = 'cpu'
        return self
    
    def unpack_group(self, group):
        for k, v in getattr(self, '{}_group'.format(self.which)).items():
            setattr(self, k, None)
            
        for k, v in group.items():
            if not hasattr(self, k) or getattr(self, k) is None:
                setattr(self, k, v)
    
    def train(self):
        """
        Switch to training dataset.
        """
        self.unpack_group(self.train_group)
        self.which='train'
        return self
        
    def valid(self):
        """
        Switch to validation dataset.
        """
        self.unpack_group(self.valid_group)
        self.which='valid'
        return self
        
    def test(self):
        """
        Switch to testing dataset.
        """
        self.unpack_group(self.test_group)
        self.which='test'
        return self
     
    def dynamic_cast(self, data):
        """
        Cast data according to the current state of the class.
        E.g. when backend=='torch', device=='cuda' and the inputed
        data is numpy.ndarray, the array will be converted to
        torch.Tensor and transfered on the GPU.
        
        Parameters
        ----------
        data : numpy.ndarray or torch.Tensor
            The data to cast
            
        Returns
        -------
        casted_data : numpy.ndarray or torch.Tensor
            The casted data
        """
        if self.backend=='torch' and isinstance(data, np.ndarray):
            data = torch_cast_cpu(data)
        if self.backend=='numpy' and isinstance(data, torch.Tensor):
            data = data.cpu().numpy()
        if isinstance(data, torch.Tensor):
            if self.device=='cuda' and data.device.type=='cpu':
                data = data.cuda()
            if self.device=='cpu' and data.device.type=='cuda':
                data = data.cpu()
        return data
            
    def __repr__(self):
        s = self.rng.get_state()
        r = 'Loader: {} [{} ({})]\nSet: {}\nRNG\'s State (hash): {}'
        return r.format(
            self.__class__.__name__,
            self.backend,
            self.device,
            self.which,
            self.rng.get_state_hash()
        )
    
    def get_state(self):
        return self.rng.get_state()
    
    def set_state(self, state):
        #fork rng before updating since it might be use elsewhere (e.g. the global np.random)
        #for using the same rng across multiple objects, use set_rng
        rng = np.random.RandomState()
        rng.set_state(state)
        self.set_rng(rng)
        return self
    
    #for easy sharing rng across multiple object
    def get_rng(self):
        """
        Utility method to get the rng.
        
        Returns
        -------
        rng : numpy.random.RandomState
            The rng used inside the utility class TorchNumpyRNG.
        """
        return self.rng.rng
    
    def set_rng(self, rng):
        """
        Utility method to set the rng.
        
        Parameters
        ----------
        rng : numpy.random.RandomState or TorchNumpyRNG
            The rng to use going forward.
        
        Returns
        -------
        self : Loader
        """
        if isinstance(rng, TorchNumpyRNG):
            #rewrap for consistency
            self.rng = TorchNumpyRNG(rng.rng)
        else:
            self.rng = TorchNumpyRNG(rng)
        
        #casting the numpy-cpu rng if necessary
        if self.backend=='torch':
            self.rng.torch()
        if self.device=='cuda':
            self.rng.cuda()
        return self
    
class IRLoader(Loader):
    """
    A subclass of Loader meant for Information Retrieval. This introduces the
    notion of mode which gouverns to way batches will be given.
    
    Parameters
    ----------
    mode : str
        should be in IRLoader.get_available_modes()
    """
    def __init__(self, mode, which, backend, device, rng=None):
        super().__init__(which, backend, device, rng=rng)
        self.value_check('mode', self.get_available_modes(), mode)
        self.mode = mode
        
    def __repr__(self):
        s = self.rng.get_state()
        r = 'Loader: {} [{} ({})]\nSet: {}\nBatch Mode: {}\nRNG\'s State (hash): {}'
        return r.format(
            self.__class__.__name__,
            self.backend,
            self.device,
            self.which,
            self.mode,
            self.rng.get_state_hash()
        )
    
    def get_available_modes(self):
        return {
            'unsupervised',
            'class_relation',
            'relational_triplets',
            'relational_matrix'
        }


import torch
import numpy as np
from torch.nn.utils import clip_grad_value_
from radbm.search.elba import EfficientLearnableBinaryAccess
softplus = torch.nn.Softplus()

class HashNet(EfficientLearnableBinaryAccess):
    """
    HashNet as in "HashNet: Deep Learning to Hash by Continuation"
    by Zhangjie Cao, Mingsheng Long, Jianmin Wang and Philip S. Yu.

    Parameters
    ----------
    fq : torch.nn.Module
        The query Multi-Bernoulli encoder.
    fd : torch.nn.Module
        The document Multi-Bernoulli encoder.
    struct : BaseSDS subclass
        The structure used in ELBA.
    match_prob : float (in [0,1])
        The probability that there is a match given a random query
        and a random document.
    alpha : float (optional)
        The HashNet's alpha used in the adaptative sigmoid (default 1)
    beta : float (optional)
        The HashNet's beta used for ramping the stage in the tanh(stage*beta*x).
        (default 1)
    """
    def __init__(self, fq, fd, struct, match_prob, alpha=1, beta=1, *args, **kwargs):
        super().__init__(fq, fd, struct, *args, **kwargs)
        self.match_prob = match_prob
        self.alpha = alpha
        self.beta = beta
        
    def step(self, q, d, r, stage=1):
        """
        Do a training step.
        
        Parameters
        ----------
        q : torch.Tensor
            A batch of queries.
        d : torch.Tensor
            A batch of documents.
        r : torch.Tensor (dtype=torch.bool)
            A matrix (2D tensor) with r[i,j] indicating if q[i] match with d[j]
        stage : float (optional)
            The stage for computing the tanh(stage*beta*x). (default=1)
            
        Returns
        -------
        loss : torch.Tensor (size 1)
            The loss of HashNet.
        """
        self.zero_grad()
        zq = self.fq(q)
        zd = self.fd(d)
        bq = torch.tanh(stage*self.beta*zq)
        bd = torch.tanh(stage*self.beta*zd)
        sh = (bq[:,None]*bd[None,:]).sum(dim=2)
        ash = self.alpha*sh
        p = self.match_prob
        w = (r/p + ~r/(1-p))
        losses = w*(softplus(ash) - r*ash)
        loss = losses.mean()
        loss.backward()
        clip_grad_value_(self.parameters(), 5)
        self.optim.step()
        return loss


import unittest
import numpy as np
from radbm.search.base import BaseSDS

class SingleDummy(BaseSDS):
    def __init__(self):
        self.index = dict()
    
    def insert(self, document, index):
        td = tuple(document)
        if td in self.index: self.index[td].add(index)
        else: self.index[td] = {index}
    
    def search(self, query):
        return self.index[tuple(query)]
    
    def itersearch(self, query):
        #yield everything
        yield from self.index.values()
    
class BatchDummy(BaseSDS):
    def __init__(self):
        self.index = dict()
    
    def batch_insert(self, documents, indexes):
        for d, i in zip(documents, indexes):
            td = tuple(d)
            if td in self.index: self.index[td].add(i)
            else: self.index[td] = {i}
    
    def batch_search(self, queries):
        return [self.index[tuple(query)] for query in queries]
    
    def batch_itersearch(self, queries):
        #yield everything for every queries
        return [self.index.values() for query in queries]
    
class TestBaseSDS(unittest.TestCase):
    def test_NotImplementedError(self):
        dummy = BaseSDS()
        with self.assertRaises(NotImplementedError):
            dummy.insert(123, 123)
        with self.assertRaises(NotImplementedError):
            dummy.batch_insert(123, 123)
        with self.assertRaises(NotImplementedError):
            dummy.search(123)
        with self.assertRaises(NotImplementedError):
            dummy.batch_search(123)
        with self.assertRaises(NotImplementedError):
            next(dummy.itersearch(123))
        with self.assertRaises(NotImplementedError):
            dummy.batch_itersearch(123)
            
    def test_default_single(self):
        dummy = BatchDummy()
        dummy_data = np.zeros((100,))
        dummy.insert(dummy_data, 0)
        out = dummy.search(dummy_data)
        self.assertEqual(out, {0})
        out = list(dummy.itersearch(dummy_data))
        self.assertEqual(out, [{0}])
        
    def test_default_batch(self):
        dummy = SingleDummy()
        dummy_data = np.arange(32*100).reshape(32,100)
        dummy.batch_insert(dummy_data, range(32))
        out = dummy.batch_search(dummy_data)
        self.assertEqual(out, [{i} for i in range(32)])
        out = [list(g) for g in dummy.batch_itersearch(dummy_data)]
        self.assertEqual(out, [[{i} for i in range(32)] for _ in range(32)])


import heapq
import numpy as np
from radbm.search.base import BaseSDS

class GridSearch(BaseSDS):
    """
    An exact symmetric search (query and documents are in the same space)
    GridSearch separates the space into a grid and fill each cell with
    documents using a dictionary (cell -> documents). For retrieval, the
    the cells are searched in order of distance and document are yielded
    when it is the closest found so far and when the next nearest cell is
    farther than the document (this is certifying there is no document
    closer).
    
    Parameters
    ----------
    bounds : np.ndarray
        shape=(2,d) with d the dimension of the data (query/documents)
        min_bounds, max_bounds = bounds are the bounds of the grid.
    divisions : np.ndarray, dtype=int
        shape=(d,) with d the dimension of the data (query/documents)
        the number of cell along each axis. 
    """
    def __init__(self, bounds, divisions):
        self.bounds = bounds
        self.min_bounds, self.max_bounds = bounds
        self.delta_bounds = self.max_bounds - self.min_bounds
        self.divisions = divisions
        self.grid = dict()
        
        #used in multiple places, a range of a dimensions
        self.range = np.arange(len(divisions))
        
        #width2 is used to compute L2 given the scalled data
        self.width2 = (self.delta_bounds/self.divisions)**2
        
        #directions are used for the graph search version of the cell generator
        eye = np.eye(len(divisions))
        self.directions = np.concatenate([eye, -eye])

    def compute_scaled_points(self, points):
        """
        Rescale point(s)
        
        Parameters
        ----------
        points : np.ndarray
            A point or a batch of points (query/document)
        
        Returns
        -------
        rescalled_points : np.ndarray
            rescalled_point[i] in [0, self.divisions[i]]
        """
        return self.divisions*(points - self.min_bounds)/self.delta_bounds

    def compute_cells(self, points):
        """
        Returns
        -------
        cell : np.ndarray, dtype=int
            The cells of the corresponding points referended by the
            closest to the origin corner.
        """
        scaled_points = self.compute_scaled_points(points)
        return np.floor(scaled_points).astype(np.int)

    def compute_cell_relative_positions(self, points):
        """
        Gives the position relative to the cell of the points
        Returns
        -------
        cell_pos : np.ndarray
            cell_pos in [0, 1]
        """
        scaled_points = self.compute_scaled_points(points)
        cells = np.floor(scaled_points).astype(np.int)
        return scaled_points - cells

    def is_cell_inbound(self, cell):
        return all(0 <= c < d for c,d in zip(cell, self.divisions))
    
    def is_points_inbound(self, points):
        inside = np.logical_and(self.min_bounds <= points, points < self.max_bounds)
        return np.all(inside)
    
    def batch_insert(self, points, indexes):
        if not self.is_points_inbound(points):
            raise ValueError('points out of bounds')
            
        #grouping points by cell
        grid = dict()
        cells = self.compute_cells(points)
        for cell, point, index in zip(cells, points, indexes):
            tcell = tuple(cell)
            if tcell in grid:
                grid[tcell][0].append(point)
                grid[tcell][1].append(index)
            else: grid[tcell] = [[point], [index]]
                
        #updating self.grid
        for tcell, (new_points, new_indexes) in grid.items():
            if tcell in self.grid:
                points, indexes = self.grid[tcell]
                self.grid[tcell] = [
                    np.concatenate([points, new_points]),
                    np.concatenate([indexes, new_indexes]),
                ]
            else: 
                self.grid[tcell] = [
                    np.array(new_points),
                    np.array(new_indexes)
                ]
        
        return self

    def point_relative_cells_dists(self, delta_x, relative_cells):
        signature = (0<relative_cells).astype(int)
        a = delta_x[signature, self.range]
        mask = (0!=relative_cells)
        tmp = (a + (np.abs(relative_cells)-1))**2 * self.width2
        return np.sqrt((tmp*mask).sum(axis=1))

    def graphsearch_cell_generator(self, point):
        scaled_point = self.compute_scaled_points(point)
        #zero_cell = compute_cells(point)
        zero_cell = np.floor(scaled_point).astype(np.int)
        #zero_cell_pos = compute_cell_relative_positions(point)
        zero_cell_pos = scaled_point - zero_cell
        delta_x = np.stack([zero_cell_pos, 1 - zero_cell_pos])
        tzero_cell =  tuple(zero_cell)
        zero = tuple(0 for _ in self.range)
        heap = [(0, zero)]
        expanded_set = {zero}
        while heap:
            dist, tcell = heapq.heappop(heap)
            yield dist, tuple((zero_cell+tcell).astype(int))
            ncells = self.directions + tcell
            dists = self.point_relative_cells_dists(delta_x, ncells)
            for ncell, dist in zip(ncells, dists):
                tncell = tuple(ncell)
                if tncell not in expanded_set and self.is_cell_inbound(tncell+zero_cell):
                    expanded_set.add(tncell)
                    heapq.heappush(heap, (dist, tncell))

    def compute_cells_dist(self, point):
        pos = self.compute_scaled_points(point)
        cell = pos.astype(int)
        cell_pos = pos - cell
        ndim = len(self.divisions)
        dists = np.array(0).reshape(*(1 for dim in range(ndim)))
        for dim, div, w2, k, d in zip(range(ndim), self.divisions, self.width2, cell, cell_pos):
            right = [i+(1-d) for i in range(div-k-1)]
            left = [(k-i-1)+d for i in range(k)]
            view = tuple(1 if i!=dim else -1 for i in range(ndim))
            axis_dists = np.concatenate([left, [0], right])**2
            dists = dists + axis_dists.reshape(view)*w2
        return np.sqrt(dists)

    def bruteforce_cell_generator(self, point):
        dists = self.compute_cells_dist(point)
        flatten_dists = dists.flatten()
        #mlog(m) with m the number of cell
        argsort = np.argsort(flatten_dists)
        sorted_dists = flatten_dists[argsort]
        sorted_cells = np.stack(np.unravel_index(argsort, self.divisions), axis=1)
        for dist, cell in zip(sorted_dists, sorted_cells):
            yield dist, tuple(cell)

    def itersearch(self, point, cell_generator='bruteforce',
                   max_dist=np.inf, max_items=np.inf,
                   yield_point=False, yield_dist=False):
        """
        Generator
        
        Parameters
        ----------
        point : np.ndarray
        cell_generator : str
            should be bruteforce or graphsearch
        max_dist : number
        max_items : number
        yield_point : bool
        yield_dist : bool
        
        Yields
        ------
        index : int
            if yield_point is False and yield_dist is False
        point : np.ndarray
            if yield_point is True and yield_dist is False
        (dist, index) : (float, int)
            if yield_point is False and yield_dist is True
        (dist, point) : (float, np.ndarray)
            if yield_point is True and yield_dist is True
        """
        if point.ndim != 1:
            raise ValueError('expected flatt array, got ndim={}'.format(point.ndim))
        
        if not self.is_points_inbound(point):
            raise ValueError('point out of bounds')
            
        if cell_generator == 'bruteforce':
            cell_generator = self.bruteforce_cell_generator(point)
        elif cell_generator == 'graphsearch':
            cell_generator = self.graphsearch_cell_generator(point)
        else:
            msg = 'cell_generator must be bruteforce or graphsearch, got {}'
            raise ValueError(msg.format(cell_generator))
                    
        if max_items <= 0 or max_dist < 0:
            #nothing to do
            return
        
        heap = []
        item_count = 0
        for dist, cell in cell_generator:
            while heap and heap[0][0] < dist and heap[0][0] < max_dist:
                d, p, i = heapq.heappop(heap)
                x = p if yield_point else i
                yield (d, x) if yield_dist else x
                item_count += 1
                if max_items <= item_count:
                    #yielded enough items
                    return   
            if max_dist < dist:
                #all further cells are farther
                #we already yielded every documents
                #closer than max_dist
                return
            if cell in self.grid:
                points, indexes = self.grid[cell]
                dists = np.sqrt(((point-points)**2).sum(axis=1))
                for d, p, i in zip(dists, points, indexes):
                    #the next lines are for debugging
                    #if d < dist:
                    #    msg = (
                    #        'something went wrong, found a point {} at '
                    #        'distance {:.4f} in the cell {} at distance {:.4f}'
                    #    )
                    #    raise ValueError(msg.format(p, d, cell, dist))
                    heapq.heappush(heap, (d, p, i))
        #empty the heap
        while heap and heap[0][0] < max_dist:
            d, p, i = heapq.heappop(heap)
            x = p if yield_point else i
            yield (d, x) if yield_dist else x
            item_count += 1
            if max_items <= item_count:
                #yielded enough items
                return


import unittest
from time import time, sleep

from radbm.utils.time import *

class TestChronometer(unittest.TestCase):
    def setUp(self):
        self.chrono = Chronometer()
            
    def test_basic_timing(self):
        self.chrono.start()
        sleep(.1)
        self.chrono.stop()
        err = abs(self.chrono.time() - .1)
        self.assertTrue(err < .001)
    
        #waste time when stopped
        sleep(.5)

        #test time addition (second start) and not stopped time() call
        self.chrono.start()
        sleep(.2)
        err = abs(self.chrono.time() - .3)
        self.assertTrue(err < .001)
        
        #test second timing without stop 
        sleep(.25)
        err = abs(self.chrono.time() - .55)
        self.assertTrue(err < .002)
        
    def test_reset(self):
        self.chrono.start()
        sleep(0.1)
        self.chrono.reset()
        self.assertTrue(self.chrono.time()==0)
        
    def test_double_start_error(self):
        self.chrono.start()
        with self.assertRaises(ValueError):
            self.chrono.start()
        
    def test_double_stop_error(self):
        with self.assertRaises(ValueError):
            self.chrono.stop()
            
    def test_repr(self):
        self.assertTrue(repr(self.chrono.start()).endswith('(start)'))
        self.assertTrue(repr(self.chrono.stop()).endswith('(stop)'))


import unittest
import numpy as np

from radbm.metrics.oracle import *
rng = np.random.RandomState(0xcafe)

#fake model
class Model(object):
    def __init__(self, N, relevant_sets, precision):
        self.N = N
        self.relevant_sets = relevant_sets
        self.precision = precision #precision
        
    def candidates(self, q):
        rs = iter(self.relevant_sets[q])
        for i in range(self.N+1):
            if rng.uniform(0,1) < self.precision:
                try: yield [next(rs)]
                except StopIteration: break
            else: yield [self.N+1] #not in relevant_set

class TestOracleMetric(unittest.TestCase):
    def query_iterator(self, M, relevant_sets):
        for i in range(M):
            yield i, relevant_sets[i]
            
    def test_basic_oracle_metric(self):
        N = 10000
        M = 100
        k = 5
        relevant_sets = [set(rng.randint(0,N,k)) for _ in range(M)]
        model_precision = .7
        query_iterator = self.query_iterator(M, relevant_sets)
        model = Model(N, relevant_sets, model_precision)
        ToCRs = time_over_oracle_calls_reduction_metric(
            model, query_iterator, N, recalls=[.1,.25,.5,1])
        
    def test_bad_candidates_oracle_metric(self):
        N = 10000
        M = 100
        k = 5
        relevant_sets = [set(rng.randint(0,N,k)) for _ in range(M)]
        model_precision = .00001 #the model will fail to achieve 1 recall
        query_iterator = self.query_iterator(M, relevant_sets)
        model = Model(N, relevant_sets, model_precision)
        with self.assertRaises(RuntimeError):
            ToCRs = time_over_oracle_calls_reduction_metric(
                model, query_iterator, N, recalls=[.1,.25,.5,1])



import heapq
import numpy as np

class _Subset(object):
    def __init__(self, items, total):
        self.items = items
        self.total = total
    
    def __hash__(self):
        return hash(tuple(self.items))
    
    def __eq__(self, subset):
        return self.items == subset.items
    
    def __lt__(self, subset):
        return self.total < subset.total
    
def _next_subsets_generator(items, total, values):
    if not items or items[0] != 0:
        yield _Subset([0]+items, total + values[0])
    for n, i in enumerate(items[:-1]):
        if i+1 < items[n+1]:
            yield _Subset(
                items[:n]+[i+1]+items[n+1:],
                total - values[i] + values[i+1]
            )
    if items:
        i = items[-1]
        if i+1 < len(values):
            yield _Subset(
                items[:-1]+[i+1],
                total - values[i] + values[i+1]
            )
        
def _least_k_subsortedset_sum_generator(values, k=None):
    if any(v < 0 for v in values): raise ValueError('values must be positive')
    if k is None: k = 2**len(values)
    k = min(k, 2**len(values))
    heap = [_Subset([], 0)]
    produced = set()
    for _ in range(k):
        subset = heapq.heappop(heap)
        yield subset.items
        for new_subset in _next_subsets_generator(subset.items, subset.total, values):
            if new_subset not in produced:
                heapq.heappush(heap, new_subset)
                produced.add(new_subset)

def least_k_subset_sum_generator(values, k=None):
    """
    Generator that yields the subset of index of values in
    increasing order of their sum. The values must be all positive.
    
    Parameters
    ----------
    values : numpy.ndarray
        The values from which to take the subsets
        
    Yields
    ------
    subset : tuple
        Subset of index of the values in increasing order of their sum
    """
    perm = list(np.argsort(values))
    values = values[perm]
    for items in _least_k_subsortedset_sum_generator(values, k):
        yield tuple(perm[i] for i in items)
        
def greatest_k_multi_bernoulli_outcomes_generator(log_probs0, log_probs1=None, k=None):
    """
    Generator that yields the outcomes of a Multi-Bernoulli in decreasing
    order of probability. This work by reducing to the problem of generating
    the subset of a set in increasing order of their sum.
    
    Notes
    -----
    Bits probability must be in ]0,1[ (i.e. they cannot be zero or one)
    
    For numerical stability, it is possible to provide 
    
    Parameters
    ----------
    log_probs0 : numpy.ndarray
        log probabilities for bits to be zero
    log_probs1 : numpy.ndarray, optional
        log probabilities for bits to be one if not given
        log_probs1 = np.log(1-np.exp(log_probs0)) which might be unstable
    k : int, optional
        The maximum number of outcomes to yield, by default all outcomes are yielded
    
    Yields
    ------
        bits : tuple
            The bits outcomes in decreasing order of prabability
    """
    if log_probs1 is None:
        log_probs1 = np.log(1-np.exp(log_probs0))
    diff = log_probs1 - log_probs0
    most_probable_outcome = (0 < diff) + 0
    flipped_diff = np.where(most_probable_outcome, diff, -diff)
    for subset in least_k_subset_sum_generator(flipped_diff, k):
        subset = list(subset)
        bits = most_probable_outcome.copy()
        bits[subset] = 1 - bits[subset]
        yield tuple(bits)


import torch
from torch.nn.utils import clip_grad_value_
import numpy as np

from radbm.utils import Ramp
from radbm.search.elba import EfficientLearnableBinaryAccess
from radbm.utils.torch import (
    torch_lme,
    torch_lse,
    positive_loss_adaptative_l2_reg,
    multi_bernoulli_equality,
    log_poisson_binomial,
    torch_log_prob_any,
)

def _importance(match, match_prob):
    batch_prob = match.float().mean()
    r1 = match_prob/batch_prob
    r0 = (1-match_prob)/(1-batch_prob)
    return torch.log(r1*match + r0*~match)

def _gamma(beta, log_match_prob, log_m):
    b2 = beta**2
    log_b2 = np.log(b2)
    c1 = np.log(1+b2) - log_b2
    c2 = log_match_prob + log_b2
    return c1 - torch.nn.Softplus()(log_m-c2)

def fbeta_loss(match, log_match, beta, log_match_prob):
    """
    Compute the Fbeta Loss
    
    Parameters
    ----------
    match : torch.Tensor (dtype=torch.bool)
        Indicate if there is a match or not
    log_match : torch.Tensor (dtype=torch.float)
        The log probability of matching given by an estimator
    beta : float
        The beta in the Fbeta Loss
    log_match_prob : float
        The log probability that there is a match given a random query
        and a random document.
        
    Returns
    -------
    fbeta_loss : torch.Tensor (dtype=torch.float)
        Containing 1 element, this is the Fbeta Loss of the estimator w.r.t match.
    """
    log_r_hats = log_match[match]
    log_m_hats = log_match + _importance(match, np.exp(log_match_prob))
    log_rho = log_r_hats.mean()
    log_m_hat = torch_lme(log_m_hats)
    log_g = _gamma(beta, log_match_prob, log_m_hat)
    log_fbeta = log_rho + log_g
    return -log_fbeta

class Fbeta(EfficientLearnableBinaryAccess):
    """
    The Fbeta model is an ELBA that use the Fbeta Loss to train the Multi-Bernoulli encoders.
    
    Parameters
    ----------
    fq : torch.nn.Module
        The query Multi-Bernoulli encoder.
    fd : torch.nn.Module
        The document Multi-Bernoulli encoder.
    struct : BaseSDS subclass
        The structure used in ELBA.
    log_match_prob : float
        The log probability that there is a match given a random query
        and a random document.
    sim : function (torch.Tensor \times torch.Tensor -> torch.Tensor)
        A function taking the query's Multi-Bernoulli code and the document's
        Multi-Bernoulli code and return the bitwise log probability that each bit match.
    match_dist : int (optional)
        The maximum number of dissimilar bits for which we consider that two binary
        vectors are matching. (default=0)
    nindex : int (optional)
        The number of indexes. (default=1)
    ramp : function (optional)
        A ramping function used for ramping the log2(beta) at each steps.
    """
    def __init__(self, fq, fd, struct, log_match_prob, sim=multi_bernoulli_equality, match_dist=0, nindex=1, ramp=Ramp(0,1024,-64,-8)):
        super().__init__(fq, fd, struct)
        self.sim = sim
        self.ramp = (lambda x: ramp) if isinstance(ramp, (int, float)) else ramp
        self.log_match_prob = log_match_prob
        self.match_dist = match_dist
        self.nindex = nindex
        
    def _loss(self, match, log_match, nbatch=None):
        log2_beta = self.ramp(nbatch)
        loss = fbeta_loss(match, log_match, 2**log2_beta, self.log_match_prob)
        return loss
    
    def index_view(self, x):
        bs, total_bits = x.shape
        if total_bits%self.nindex!=0:
            msg = 'The total numbers of bit should be divisible by the number of indexes. ({}%{} != 0)'
            raise ValueError(msg.format(total_bits, self.nindex))
        bits_per_index = total_bits//self.nindex
        return x.view(bs, self.nindex, bits_per_index)
    
    def loss(self, q, d, match, l2_ratio=0.01, nbatch=None):
        zq = self.index_view(self.fq(q))
        zd = self.index_view(self.fd(d))
        log_p0, log_p1 = self.sim(zq[:,None], zd[None,:])
        log_pb = log_poisson_binomial(log_p1, log_p0) #inverting log_p0 and log_p1 to count the zeros
        log_index1 = torch_lse(log_pb[..., :self.match_dist+1], dim=3)
        log_index0 = torch_lse(log_pb[..., self.match_dist+1:], dim=3)
        _, log_match = torch_log_prob_any(log_index0, log_index1)
        loss = self._loss(match, log_match, nbatch)
        loss = positive_loss_adaptative_l2_reg(loss, l2_ratio, [zq,zd])
        return loss
        
    def step(self, q, d, match, l2_ratio=0.01, nbatch=None):
        """
        Do a training step.
        
        Parameters
        ----------
        q : torch.Tensor
            A batch of queries.
        d : torch.Tensor
            A batch of documents.
        match : torch.Tensor (dtype=torch.bool)
            A matrix (2D tensor) with match[i,j] indicating if q[i] match with d[j]
        l2_ratio : float (optional)
            The wanted ratio between the Fbeta Loss and L2 regularization. (default 0.01)
        nbatch : int or None (optional)
            Give the number of batch, this is uses for ramping. If None, this uses the
            final ramping value.
            
        Returns
        -------
        loss : torch.Tensor (size 1)
            The loss of the current batch.
        """
        self.zero_grad()
        loss = self.loss(q, d, match, l2_ratio=l2_ratio, nbatch=nbatch)
        loss.backward()
        clip_grad_value_(self.parameters(), 5)
        self.optim.step()
        return loss


import torch
import numpy as np
from functools import reduce
from collections import OrderedDict

def torch_hamming(code1, code2, dim=-1, keepdims=False):
    #obsolete, a way better implementation in radbm.metrics.hamming
    if isinstance(dim, int): dim = (dim,)
    n1 = int(np.prod([code1.shape[d] for d in dim]))
    n2 = int(np.prod([code2.shape[d] for d in dim]))
    if n1 != n2: raise ValueError()
    c = (code1==code2).sum(dim=dim, keepdims=keepdims)
    return n1 - c

def torch_soft_hamming(v, w, dim=-1, keepdims=False):
    if isinstance(dim, int): dim = (dim,)
    n1 = int(np.prod([v.shape[d] for d in dim]))
    n2 = int(np.prod([w.shape[d] for d in dim]))
    if n1 != n2: raise ValueError()
    dot = (v*w).sum(dim=dim, keepdims=keepdims)
    return .5*(n1 - dot)

def torch_maximum(*tensors):
    return reduce(torch.max, tensors)

def torch_logsumexp(*tensors):
    maxes = torch_maximum(*tensors)
    maxes = torch.where(maxes==-np.inf, torch.zeros_like(maxes), maxes)
    return sum((tensor - maxes).exp() for tensor in tensors).log() + maxes

def torch_logsubexp(log_a, log_b):
    """
    Compute log(a-b) = log(exp(log(a))-exp(log(b)))
    This is unsound if a < b (log_a < log_b)
    The only thing this function does is removing
    the risk of overflow in the exponentials.
    log(a-b) = log(exp(log_a)*(1 - exp(log_b-log_a))) = log(1-exp(log_b-log_a)) + log_a
    """
    where = log_a <= log_b
    if torch.all(where):
        n = where.sum()
        total = np.prod(where.shape)
        msg = 'Cannot compute the log of the a-b if it is negative, {}/{} negative values found'
        raise ValueError(msg.format(n, total))
    return torch.log(1-torch.exp(log_b-log_a)) + log_a

def torch_max(tensor, dim=None, keepdims=False):
    if isinstance(dim, int):
        return tensor.max(dim, keepdim=keepdims)[0]
    if dim is None or set(dim)==set(range(tensor.dim())):
        if keepdims: return tensor.max().view(*[1]*tensor.dim())
        else: return tensor.max()
    dim = tuple(d+tensor.dim() if d < 0 else d for d in dim)
    for d in sorted(dim, reverse=True):
        tensor = tensor.max(d, keepdim=keepdims)[0]
    return tensor

def torch_lse(tensor, dim=None, keepdims=False):
    if dim is None: dim = tuple(range(tensor.dim()))
    maxes = torch_max(tensor, dim=dim, keepdims=True)
    maxes = torch.where(maxes==-np.inf, torch.zeros_like(maxes), maxes)
    lse = (tensor - maxes).exp().sum(dim=dim, keepdim=True).log() + maxes
    if not keepdims:
        if isinstance(dim, (tuple,list)):
            dim = tuple(d+tensor.dim() if d < 0 else d for d in dim)
            for d in sorted(dim, reverse=True): lse = lse.squeeze(dim=d)
        else: lse = lse.squeeze(dim=dim)
    return lse

def torch_lme(tensor, dim=None, keepdims=False):
    #log mean exp
    return torch_lse(tensor, dim, keepdims) - np.sum(np.log(tensor.shape))

def torch_categorical_entropy(log_p, dim=-1, keepdims=False):
    log_p = torch.where(log_p==-np.inf, torch.zeros_like(log_p), log_p)
    return -(log_p.exp()*log_p).sum(dim=dim, keepdims=keepdims)

def torch_multibernoulli_entropy(log_p0, log_p1, dim=-1, keepdims=False):
    log_p0 = torch.where(log_p0==-np.inf, torch.zeros_like(log_p0), log_p0)
    log_p1 = torch.where(log_p1==-np.inf, torch.zeros_like(log_p1), log_p1)
    return -(log_p0.exp()*log_p0 +log_p1.exp()*log_p1).sum(dim=dim, keepdims=keepdims)
    
def params_to_buffer(module):
    module._buffers.update(module._parameters)
    module._parameters = OrderedDict()
    for child in module._modules.values():
        params_to_buffer(child)
        
def buffer_no_grad(module):
    for p in module.buffers():
        p.requires_grad = False
    for child in module._modules.values():
        buffer_no_grad(child)

def _fbeta_gamma_term(beta, log_q, log_s):
    b2 = beta**2
    log_b2 = np.log(b2)
    c1 = np.log(1+b2) - log_b2
    c2 = log_q + log_b2
    return c1 - torch.nn.Softplus()(log_s-c2)

def torch_negative_log_fbeta_loss(log_r_hats, log_s_hats, beta, log_q):
    log_rho = log_r_hats.mean()
    log_s_hat = torch_lme(log_s_hats)
    log_g = _fbeta_gamma_term(beta, log_q, log_s_hat)
    log_fbeta = log_rho + log_g
    return -log_fbeta

def positive_loss_adaptative_l2_reg(loss, ratio, values):
    #model bad -> high loss -> high regularization
    #model good -> low loss -> low regularization
    reg = sum((v**2).sum() for v in values)
    alpha = ratio*float(loss/reg)
    return loss + alpha*reg

def torch_complex_polar(a, b):
    #return r, theta
    r = torch.sqrt(a**2 + b**2)
    theta = torch.atan2(b, a)
    return r, theta

def torch_complex_log(a, b):
    r, theta = torch_complex_polar(a, b)
    return r.log(), theta #<-- cartesian coordinate of log(a+ib) = log(r) + i*theta

def torch_complex_exp(a, b):
    exp_a = a.exp()
    return exp_a*torch.cos(b), exp_a*torch.sin(b) #Euler

def torch_complex_prod(a, b, dim=None):
    #use log space because addition of complex numbers is elementwise 
    #(i.e. real with real and complex with complex)
    if dim is None: dim = tuple(range(len(a.shape)))
    a, b = torch_complex_log(a, b)
    return torch_complex_exp(a.sum(dim=dim), b.sum(dim=dim))

class PoissonBinomialSum(torch.nn.Module):
    def __init__(self, n, roots_requires_grad=False):
        super().__init__()
        self.n = n
        w = 2*np.pi/(n+1)
        #use python imaginary arithmetic
        roots = np.array([np.e**(w*l*1j) for l in range(n+1)])
        roots_real = torch.tensor(roots.real, requires_grad=roots_requires_grad)
        roots_imag = torch.tensor(roots.imag, requires_grad=roots_requires_grad)
        self.register_buffer('roots_real', roots_real)
        self.register_buffer('roots_imag', roots_imag)
        
    def forward(self, q):
        q_view = q.view(1,self.n)
        z_real = self.roots_real.view(self.n+1,1)*q_view - q_view + 1
        z_imag = self.roots_imag.view(self.n+1,1)*q_view
        x_real, x_imag = torch_complex_prod(z_real, z_imag, dim=1)
        x = torch.stack([x_real, x_imag], dim=1)
        pb = torch.fft(x,1)
        return pb[:,0]/(self.n+1)
    
def torch_cast_cpu(x, max_floating_point=32):
    if not isinstance(x, np.ndarray): return torch.tensor(x)
    dtype = x.dtype.type
    if max_floating_point == 64 and issubclass(dtype, np.float128):
        return torch.tensor(x, dtype=torch.float64)
    if max_floating_point == 32 and issubclass(dtype, (np.float128, np.float64)):
        return torch.tensor(x, dtype=torch.float32)
    if max_floating_point == 16 and issubclass(dtype, (np.float128, np.float64, np.float32)):
        return torch.tensor(x, dtype=torch.float16)
    return torch.tensor(x)

def torch_cast(x, max_floating_point=32):
    x = torch_cast_cpu(x, max_floating_point=max_floating_point)
    if torch.cuda.is_available(): x = x.cuda()
    return x

class TorchNumpyRNG(object):
    def __init__(self, rng=np.random):
        self.rng=rng
        self.device='cpu'
        self.backend='numpy'
        self.numpy()
        
    def apply_switch(self, wrap):
        bypass = {'seed', 'get_state', 'set_state'}
        for attr in np.random.__all__:
            if attr in bypass: setattr(self, attr, getattr(self.rng, attr))
            elif hasattr(self.rng, attr) and callable(getattr(self.rng, attr)):
                setattr(self, attr, wrap(getattr(self.rng, attr)))
    
    def torch_cpu_wrap(self, f):
        def wrapped(*args, **kwargs):
            a = f(*args, **kwargs)
            return torch_cast_cpu(a)
        return wrapped
    
    def torch_cuda_wrap(self, f):
        def wrapped(*args, **kwargs):
            a = f(*args, **kwargs)
            return torch_cast(a)
        return wrapped
    
    def numpy_cpu_wrap(self, f):
        return f
    
    def torch(self):
        if self.device=='cpu':
            self.apply_switch(self.torch_cpu_wrap)
        elif self.device=='cuda':
            self.apply_switch(self.torch_cuda_wrap)
        self.backend = 'torch'
        return self
    
    def numpy(self):
        if self.device=='cuda':
            raise ValueError('cannot use numpy backend with cuda, use rng.cpu() first')
        self.apply_switch(self.numpy_cpu_wrap)
        self.backend = 'numpy'
        return self
    
    def cpu(self):
        if self.backend=='numpy':
            self.apply_switch(self.numpy_cpu_wrap)
        elif self.backend=='torch':
            self.apply_switch(self.torch_cpu_wrap)
        self.device = 'cpu'
        return self
    
    def cuda(self):
        if self.backend=='numpy':
            raise ValueError('cannot use cuda with numpy backend, use rng.torch() first')
        self.apply_switch(self.torch_cuda_wrap)
        self.device = 'cuda'
        return self
    
    def get_state_hash(self):
        s = self.get_state()
        #don't hash string (i.e. s[0]) since string's hash are salted at each session
        #s[0] is always 'MT19937' anyway
        return hash(tuple(s[1])+s[2:])
    
    def __repr__(self):
        s = 'TorchNumpyRNG\nBackend: [{} ({})]\nState\'s hash: {}'
        return s.format(
            self.backend,
            self.device,
            self.get_state_hash()
        )
    
class TorchNumpy(object):
    def __init__(self, data, backend, device):
        self.backend = backend
        self.device = device
        if backend not in {'numpy', 'torch'}:
            raise ValueError('backend must be numpy or torch, got {}'.format(backend))
            
        if backend=='numpy' and isinstance(data, torch.Tensor):
            data = data.detach().cpu().numpy()
        elif backend=='torch' and isinstance(data, np.ndarray):
            data = torch_cast_cpu(data)
            
        if device=='cpu': self.data = data
        elif device=='cuda':
            if backend=='torch': self.data = data.cuda()
            else: raise ValueError('cannot use cuda with numpy backend')
        else: raise ValueError('device must be cpu or cuda, got {}'.format(device))
        
    def cuda(self):
        self.data = self.data.cuda()
        
    def cpu(self):
        self.data = self.data.cpu()
        
    def torch(self):
        self.data = torch_cast_cpu(self.data)
        
    def numpy(self):
        self.data = self.data.numpy()


def unique_list(it):
    """
    Create a list from an iterable with only unique element and where
    the order is preserved.
    
    Parameters
    ----------
    it : iterable
        Items should be hashable and comparable
        
    Returns : list
        All items in the list is unique and the order of the iterable
        is preserved.
    """
    unique = set()
    return [i for i in it if i not in unique and unique.add(i) is None]

def Ramp(x0, x1, y0, y1):
    """
    Parameters
    ----------
    x0 : float
        The input value where we start ramping
    x1 : float
        The input value where we stop ramping
    y0 : float
        The output value where the ramp starts
    y1 : float
        The output value where the ramp stops
    
    Returns
    -------
    ramp : function float -> float
        The ramping function
    """
    rate = (y1-y0) / (x1-x0)
    def ramp(x):
        if x is None: return y1
        y = (x-x0)*rate + y0
        a, b = x < x0, x1 <= x
        return a*y0 + b*y1 + (1-a)*(1-b)*y
    return ramp


from .generators import (
    least_k_subset_sum_generator,
    greatest_k_multi_bernoulli_outcomes_generator,
)
from .hypergeometric import (
    hypergeometric,
    superdupergeometric,
    superdupergeometric_expectations
)


import torch

def where_update(w, v, data):
    v = torch.tensor(v, dtype=data.dtype, device=data.device)
    return torch.where(w, v, data)

def torch_rgb_to_hls(images):
    """
    Transform a batch of RGB images into HLS format.
    
    Parameters
    ----------
    images : torch.Tensor (shape (N, 3, H, W))
        The images to transform.
        
    Returns
    -------
    hls_images : torch.Tensor (shape (N, 3, H, W))
        Instead of red, green and blue there is hue, luminance and saturation.
    """
    #BS, C, H, W
    vmin, imin = images.min(dim=1)
    vmax, imax = images.max(dim=1)
    c = vmax - vmin
    zeros = c==0
    pc = where_update(zeros, 1, c)
    
    #hue
    pim = images/pc[:,None]
    r, g, b = pim[:,0], pim[:,1], pim[:,2]
    ph = (imax==0)*(g-b) + (imax==1)*(b-r+2) + (imax==2)*(r-g+4)
    h = where_update(zeros, 0, ph)/6 % 1
    
    #luminance
    l = (vmax + vmin)/2
    
    #saturation
    pl = where_update(zeros, 1, l)
    ps = c / (1 - torch.abs(2*pl - 1))
    s = where_update(zeros, 0, ps)
    
    return torch.stack([h,l,s], dim=1)

def torch_hls_to_rgb(hls_images):
    """
    Transform a batch of RGB images into HLS format.
    
    Parameters
    ----------
    hls_images : torch.Tensor (shape (N, 3, H, W))
        The images, where the channels correspond to hue, luminance and saturation.
        
    Returns
    -------
    images : torch.Tensor (shape (N, 3, H, W))
        Instead of hue, luminance and saturation there is red, green and blue.
    """
    im = hls_images
    h, l, s = im[:,[0]], im[:,[1]], im[:,[2]]
    a = s*torch.min(l, 1-l)
    h12 = 12*h
    n = torch.tensor([[[0]], [[8]], [[4]]], dtype=h12.dtype, device=h12.device)
    k = (n+h12)%12
    one = torch.tensor(1, dtype=h12.dtype, device=h12.device)
    kmin = torch.min(torch.min(k-3, 9-k), one)
    v = torch.max(-one,  kmin)
    return (l-a*torch.max(-one,  kmin)).clamp(0, 1)


import torch
import numpy as np
from scipy.stats import binom
from torch.nn.utils import clip_grad_value_
from radbm.search.elba import EfficientLearnableBinaryAccess
from radbm.utils.torch import log_hamming_binomial
log_sigmoid = torch.nn.LogSigmoid()

def kl_div(log_p, log_q, *args, **kwargs):
    return (log_p.exp()*(log_p - log_q)).sum(*args, **kwargs)

class HBKL(EfficientLearnableBinaryAccess):
    """
    Hamming Binomial with KL-divergence

    Parameters
    ----------
    fq : torch.nn.Module
        The query Multi-Bernoulli encoder.
    fd : torch.nn.Module
        The document Multi-Bernoulli encoder.
    struct : BaseSDS subclass
        The structure used in ELBA.
    nbits : int
        The number of bits
    positive_binomial_prob : float [0,1]
        The target binomial probability for matching query-document pairs
    negative_binomial_prob : float [0,1]
        The target binomial probability for non-matching query-document pairs
    """
    def __init__(self, fq, fd, struct, nbits, positive_binomial_prob, negative_binomial_prob, *args, **kwargs):
        super().__init__(fq, fd, struct, *args, **kwargs)
        self.register_buffer('pos_binom_logpmf', torch.tensor(binom(nbits, positive_binomial_prob).logpmf(range(nbits+1))))
        self.register_buffer('neg_binom_logpmf', torch.tensor(binom(nbits, negative_binomial_prob).logpmf(range(nbits+1))))
        
    def log_hamming_binomial(self, q, d):
        zq = self.fq(q)[:,None]
        zd = self.fd(d)[None,:]
        return log_hamming_binomial(*map(log_sigmoid, [-zq, zq, -zd, zd]))
    
    def loss(self, q, d, r):
        log_hb = self.log_hamming_binomial(q, d)
        pos_losses = kl_div(self.pos_binom_logpmf, log_hb[r], dim=1)
        neg_losses = kl_div(log_hb[~r], self.neg_binom_logpmf, dim=1)
        return pos_losses.mean() + neg_losses.mean() #could potentially weight those terms
        
    def step(self, q, d, r):
        """
        Do a training step.
        
        Parameters
        ----------
        q : torch.Tensor
            A batch of queries.
        d : torch.Tensor
            A batch of documents.
        r : torch.Tensor (dtype=torch.bool)
            A matrix (2D tensor) with r[i,j] indicating if q[i] match with d[j]
            
        Returns
        -------
        loss : torch.Tensor (size 1)
            The loss of HBKL.
        """
        self.zero_grad()
        loss = self.loss(q, d, r)
        loss.backward()
        clip_grad_value_(self.parameters(), 5)
        self.optim.step()
        return loss


import torch
from radbm.utils.torch import torch_lse, torch_logsumexp
from .poisson_binomial import log_poisson_binomial
logsigmoid = torch.nn.LogSigmoid()

def multi_bernoulli_equality(xz, yz):
    """
    Compute the bitwise log probability that two Multi-Bernoulli are equal.
    
    Parameters
    ----------
    xz : torch.tensor
        the logits (before sigmoid) of the first Multi-Bernoulli
    yz : torch.tensor
        the logits (before sigmoid) of the second Multi-Bernoulli
        
    Returns
    -------
    log_p0 : torch.tensor
        the bitwise log probability that the two Multi-Bernoulli are not equal
    log_p1 : torch.tensor
        the bitwise log probability that the two Multi-Bernoulli are equal
        
    Notes
    -----
    xz and yz need not to have the same shape, but they should
    be broadcastable.
    """
    xp, yp, xn, yn = map(logsigmoid, (xz, yz, -xz, -yz))
    log_p0 = torch_logsumexp(xp + yn, xn + yp)
    log_p1 = torch_logsumexp(xp + yp, xn + yn)
    return log_p0, log_p1
    
def multi_bernoulli_subset(xz, yz):
    """
    Compute the bitwise log probability that the first Multi-Bernoulli
    is lower are equal to the second.
    
    Parameters
    ----------
    xz : torch.tensor
        the logits (before sigmoid) of the first Multi-Bernoulli
    yz : torch.tensor
        the logits (before sigmoid) of the second Multi-Bernoulli
        
    Returns
    -------
    log_p0 : torch.tensor
        the bitwise log probability of not subset
    log_p1 : torch.tensor
        the bitwise log probability of subset
        
    Notes
    -----
    xz and yz need not to have the same shape, but they should
    be broadcastable.
    """
    xp, yp, xn, yn = map(logsigmoid, (xz, yz, -xz, -yz))
    log_p0 = xp + yn
    log_p1 = torch_logsumexp(xp + yp, xn + yn, xn + yp)
    return log_p0, log_p1

def multi_bernoulli_activated_equality(xz, yz, az):
    """
    Compute the bitwise log probability that two Multi-Bernoulli are equal
    or that a third Multi-Bernoulli is one.
    
    Parameters
    ----------
    xz : torch.tensor
        the logits (before sigmoid) of the first Multi-Bernoulli
    yz : torch.tensor
        the logits (before sigmoid) of the second Multi-Bernoulli
    az : torch.tensor
        the logits of the third Multi-Bernoulli which act as an activation
        of the equality.
        
    Returns
    -------
    log_p0 : torch.tensor
        the bitwise log probability that the two Multi-Bernoulli are not equal
        and the third is zero.
    log_p1 : torch.tensor
        the bitwise log probability that the two Multi-Bernoulli are equal
        or the third is one.
        
    Notes
    -----
    xz and yz need not to have the same shape, but they should
    be broadcastable.
    """
    xp, yp, ap, xn, yn, an = map(logsigmoid, (xz, yz, az, -xz, -yz, -az))
    log_p0 = torch_logsumexp(an + xp + yn, an + xn + yp)
    log_p1 = torch_logsumexp(ap, an + xp + yp, an + xn + yn)
    return log_p0, log_p1

def multi_bernoulli_activated_subset(xz, yz, az):
    """
    Compute the bitwise log probability that the first Multi-Bernoulli
    is lower are equal to the second or that a third Multi-Bernoulli is one.
    
    Parameters
    ----------
    xz : torch.tensor
        the logits (before sigmoid) of the first Multi-Bernoulli
    yz : torch.tensor
        the logits (before sigmoid) of the second Multi-Bernoulli
    az : torch.tensor
        the logits of the third Multi-Bernoulli which act as an activation
        of the "subset".
        
    Returns
    -------
    log_p0 : torch.tensor
    log_p1 : torch.tensor
        
    Notes
    -----
    xz and yz need not to have the same shape, but they should
    be broadcastable.
    """
    xp, yp, ap, xn, yn, an = map(logsigmoid, (xz, yz, az, -xz, -yz, -az))
    log_p0 = an + xp + yn
    log_p1 = torch_logsumexp(ap, an + xp + yp, an + xn + yn, an + xn + yp)
    return log_p0, log_p1

def torch_log_prob_any(log_q0, log_q1):
    """
    Similar to x.any() but for log probabilities (instead of booleans). The any is taken
    across the last dim.
    
    Parameters
    ----------
    log_q0 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be zero. The any operation is over
        the last dim. shape=(a1,a2,a3,...,am,n) where n is the number of (independant)
        Bernoullis. a1,a2,a3,...,am are arbitrary but should match with log_q1.
    
    log_q1 : torch.tensor (dtype=torch.float)
        The log probability of each bits to be one. The any operation is over
        the last dim. shape=(a1,a2,a3,...,am,n) where n is the number of (independant)
        Bernoullis. a1,a2,a3,...,am are arbitrary but should match with log_q1.
    
    Returns
    -------
    log_nor : torch.tensor (dtype=torch.float)
    log_or : torch.tensor (dtype=torch.float)
    """
    p = log_poisson_binomial(log_q0, log_q1)
    return p[..., 0], torch_lse(p[..., 1:], dim=-1)


import unittest, itertools, torch
from radbm.utils.torch import log_poisson_binomial, log_hamming_binomial

def explicit_poisson_binomial(q):
    out = torch.zeros(len(q)+1)
    for k in range(len(q)+1):
        for I in map(list, itertools.combinations(range(len(q)), k)):
            nI = [i for i in range(len(q)) if i not in I]
            out[k] += q[I].prod()*(1-q)[nI].prod()
    return out

class TestLogPoissonBinomial(unittest.TestCase):
    def test_accuracy(self):
        q = torch.tensor([0.2, 0.1, 0.8, 0.7, 0.99])
        expected = explicit_poisson_binomial(q)
        log_q0, log_q1 = (1-q).log(), q.log()
        pb = log_poisson_binomial(log_q0, log_q1).exp()
        self.assertTrue(torch.allclose(pb, expected))
        
    def test_batch(self):
        torch.manual_seed(0xcafe)
        q = torch.rand(3,4,5,64)
        log_q0, log_q1 = (1-q).log(), q.log()
        log_pb = log_poisson_binomial(log_q0, log_q1)
        log_pb123 = log_poisson_binomial(log_q0[1,2,3], log_q1[1,2,3])
        self.assertTrue(torch.allclose(log_pb[1,2,3], log_pb123))
        
    def test_shape_mismatch(self):
        q = torch.tensor([0.2, 0.1, 0.8, 0.7, 0.99])
        log_q0, log_q1 = (1-q).log(), q.log()
        with self.assertRaises(ValueError):
            log_poisson_binomial(log_q0, log_q1[:-1])
        
class TestLogHammingBinomial(unittest.TestCase):
    def test_accuracy(self):
        p1 = torch.tensor([0.2, 0.1, 0.8, 0.7, 0.99])
        p2 = torch.tensor([0.3, 0.75, 0.4, 0.5, 0.9])
        q = p1*(1-p2) + (1-p1)*p2
        expected = explicit_poisson_binomial(q)
        log_p10, log_p11 = (1-p1).log(), p1.log()
        log_p20, log_p21 = (1-p2).log(), p2.log()
        hb = log_hamming_binomial(log_p10, log_p11, log_p20, log_p21).exp()
        self.assertTrue(torch.allclose(hb, expected))


import torch
import numpy as np
from pyperlin import FractalPerlin2D
from torch.distributions.normal import Normal
from radbm.utils.torch import torch_rgb_to_hls, torch_hls_to_rgb

def squeeze_values(mu, x):
    norm1 = Normal(mu, 1)
    norm2 = Normal(-mu, 1)
    return norm2.cdf(norm1.icdf(x.clamp(0,1)))

def build_grid(N, H, W, device='cpu'):
    hlin = torch.linspace(-1, 1, H, device=device)
    wlin = torch.linspace(-1, 1, W, device=device)
    y, x = torch.meshgrid(hlin, wlin)
    return torch.stack([x, y], dim=2).expand(N, H, W, 2)

def rotation_matrices(theta):
    cos = torch.cos(theta)
    sin = torch.sin(theta)
    rotation = torch.zeros(*theta.shape, 2, 2, dtype=theta.dtype, device=theta.device)
    rotation[:,0,0] = cos
    rotation[:,0,1] = -sin
    rotation[:,1,0] = sin
    rotation[:,1,1] = cos
    return rotation

def spatial_batch_rotation(rotation, grid):
    #rotation.shape = (N, 2, 2)
    #grid.shape = (N, H, W, 2)
    return torch.matmul(rotation[:,None,None], grid[:,:,:,:,None])[:,:,:,:,0]

class AlterImages(torch.nn.Module):
    def __init__(
        self,
        shape,
        batch_global_scale_range = [1., 1.],
        hls_perlin_resolutions = [(2**i,2**i) for i in range(1,8)],
        hls_perlin_factors = [.5**i for i in range(7)],
        hls_global_noise_range = [[-.1, -.2, -.2], [.1, .2, .2]],
        hls_perlin_scale = [.5, 1, 1],
        spatial_distortion_perlin_resolutions = [(2**i,2**i) for i in range(1,4)],
        spatial_distortion_perlin_factors = [.5**i for i in range(3)],
        spatial_distortion_scale = .2,
        spatial_rotation_range = [-np.pi/16, np.pi/16],
        generator = torch.random.default_generator,
    ):
        self.shape = shape
        self.hls_perlin = FractalPerlin2D(
            shape,
            hls_perlin_resolutions,
            hls_perlin_factors,
            generator = generator
        )
        self.spatial_distortion_perlin = FractalPerlin2D(
            shape,
            spatial_distortion_perlin_resolutions,
            spatial_distortion_perlin_factors,
            generator = generator
        )
        self.device = generator.device
        self.generator = generator
        self.batch_global_scale_range = batch_global_scale_range
        self.hls_global_noise_range = torch.tensor(
            hls_global_noise_range, dtype=torch.float32, device=self.device)
        self.hls_perlin_scale = torch.tensor(
            hls_perlin_scale, dtype=torch.float32, device=self.device)[None, :, None, None]
        self.spatial_distortion_scale = spatial_distortion_scale
        self.spatial_rotation_range = spatial_rotation_range
        
    def uniform_noise(self, start, end, shape, dtype=torch.float32, device=None):
        device = self.device if device is None else device
        noise = torch.zeros(shape, dtype=dtype, device=device)
        noise.uniform_(start, end, generator=self.generator)
        return noise
    
    def alter_hls(self, images, batch_global_scale=None):
        N = len(images)
        perlin_noise = self.hls_perlin(batch_size=N*3)
        perlin_noise = perlin_noise.view(N, 3, *self.shape)#.permute(0,2,3,1)
        global_noise = self.uniform_noise(0, 1, images.shape[:2])
        global_range = self.hls_global_noise_range[1] - self.hls_global_noise_range[0]
        global_noise = global_range*global_noise + self.hls_global_noise_range[0]
        noise = self.hls_perlin_scale*perlin_noise + global_noise[:,:,None, None]
        if batch_global_scale is not None:
            noise = batch_global_scale[:,None,None,None]*noise
        images = torch_rgb_to_hls(images)
        h = (images[:,[0]] + noise[:,[0]]) % 1
        ls = squeeze_values(noise[:,1:], images[:,1:])
        images = torch.cat([h, ls], dim=1)
        return torch_hls_to_rgb(images)
    
    def alter_space(self, images, batch_global_scale=None):
        N = len(images)
        perlin_noise = self.spatial_distortion_perlin(batch_size=N*2)
        perlin_noise = perlin_noise.view(N, 2, *self.shape).permute(0,2,3,1)
        grid = build_grid(N, *self.shape, device=self.device)
        theta = self.uniform_noise(*self.spatial_rotation_range, len(images))
        if batch_global_scale is not None:
            theta = batch_global_scale*theta
            perlin_noise = batch_global_scale[:,None,None,None]*perlin_noise
        grid = spatial_batch_rotation(rotation_matrices(theta), grid)
        grid = grid + self.spatial_distortion_scale*perlin_noise
        return torch.nn.functional.grid_sample(images, grid, align_corners=False, padding_mode='zeros')
    
    def __call__(self, images):
        batch_global_scale = self.uniform_noise(*self.batch_global_scale_range, len(images))
        images = self.alter_hls(images, batch_global_scale=batch_global_scale)
        images = self.alter_space(images, batch_global_scale=batch_global_scale)
        return images


import numpy as np

def hypergeometric(N, K):
    """
    This function compute the pmf of the Hypergeometric(N, K, n)
    for each possible value of n (i.e. n in {0,1,2,...,N})
    
    In the context where an urn with N marbles contains K white marbles,
    this function outputs an array P of shape (N+1,K+1) where
    P[i, j] correspond to the probabillity that with i samples without 
    replacement we select j white marbles.
    
    Parameters
    ----------
    N : int
        The number of marbles in the urn
    K : int
        The number of white marbles in the urn
        
    Returns
    -------
    P : numpy.ndarray
        P[i, j] is the probability that with i samples without replacement
        we select j white marbles. P.shape is (N+1, K+1)
        
    Notes
    -----
    This is equivalent to 
    np.array([scipy.stats.hypergeom(N, K, n).pmf(range(0,K+1)) for n in range(N+1)])
    but faster, if only a row is needed (e.g. P[i]) than using scipy is faster.
    
    This algorithm uses the following recursive formula
    P[i, j] = P[i-1, j]*(1-Q[i-1, j]) + P[i-1, j-1]*Q[i-1, j-1]
    with Q[i, j] = (K-j)/(N-i) the probability of sampling a white marble
    given that i marbles where sampled from which j were white
    """
    P = np.zeros((N+1, K+1))
    P[0,0] = 1 #P[0, j] = 1 if j==0 else 0
    P[N,K] = 1 #P[N, j] = 1 if j==K else 0
    
    #To compute Q We don't need to bother with
    #value > 1 since they will be multiplied by 0
    Q = (K-np.arange(K+1))[None]/(N-np.arange(N))[:,None]
    
    for n in range(1,N):
        P[n] = P[n-1]*(1-Q[n-1])
        P[n,1:] += P[n-1,:-1]*Q[n-1,:-1]
        
    return P
    
def superdupergeometric(N, K):
    """
    This is the scenario where we have an urn with N marbles
    and K of them are white. We sample from the urn without replacement
    until we obtain k white marbles. The function gives the probability
    that we need n samples to obtain k white marbles.
    
    Parameters
    ----------
    N : int
        The number of marbles in the urn
    K : int
        The number of white marbles in the urn
        
    Returns
    -------
    SP : numpy.ndarray
        SP[i, j] is the probability that it requires i samples without replacement
        to select j white marbles. SP.shape is (N+1, K+1)
        
    Notes
    -----
    Probabibly related to the negative hypergeometric distribution
    
    This uses the hypergeometric (hence the name) 
    SP[i, j] = P[i-1, j-1]*Q[i-1, j-1]
    where P = hypergeometric(N, K) and with 
    Q[i, j] = (K-j)/(N-i) the probability of sampling a white marble
    given that i marbles where sampled from which j were white
    """
    
    P = hypergeometric(N, K)
    SP = np.zeros((N+1, K+1))
    SP[0, 0] = 1 #we only need 0 sample to get 0 white marbles
    
    #To compute Q We don't need to bother with
    #value > 1 since they will be multiplied by 0
    Q = (K-np.arange(K+1))[None]/(N-np.arange(N))[:,None]
    
    for n in range(1,N+1):
        SP[n,1:] += P[n-1,:-1]*Q[n-1,:-1]
        
    return SP

def superdupergeometric_expectations(N, K):
    """
    This is the scenario where we have an urn with N marbles
    and K of them are white. We sample from the urn without replacement 
    until we obtain k white marbles. The function gives the  expected
    number of samples n requires to get k white marbles (for each k).
    
    Parameters
    ----------
    N : int
        The number of marbles in the urn
    K : int
        The number of white marbles in the urn
        
    Returns
    -------
    ESP : numpy.ndarray
        ESP[k] is the expected number of samples without replacement
        requires to get k white marbles. ESP.shape is (K+1,)
        
    Notes
    -----
    This is equivalent (but way faster) to (SP*np.arange(N+1)[:,None]).sum(axis=0)
    where SP = superdupergeometric(N, K)
    """
    ESP = (N+1)/(K+1)*np.arange(K+1)
    return ESP



from .poisson_binomial import log_poisson_binomial, log_hamming_binomial
from .log_arithmetic import (
    multi_bernoulli_equality,
    multi_bernoulli_subset,
    multi_bernoulli_activated_equality,
    multi_bernoulli_activated_subset,
    torch_log_prob_any,
)


import unittest, torch
import numpy as np
from radbm.utils.torch import (
    multi_bernoulli_equality,
    multi_bernoulli_subset,
    multi_bernoulli_activated_equality,
    multi_bernoulli_activated_subset,
    torch_log_prob_any,
)

def torch_prob_any(probs):
    return 1 - (1-probs).prod(dim=-1)

class TestLogAny(unittest.TestCase):
    def test_torch_log_prob_any(self):
        gen = torch.Generator().manual_seed(0xcafe)
        probs = torch.zeros((2,3,4,32), dtype=torch.float64)
        probs.uniform_(0, .1, generator=gen)
        log_q0 = (1-probs).log()
        log_q1 = probs.log()

        log_nor, log_or = torch_log_prob_any(log_q0, log_q1)
        prob_or = torch_prob_any(probs)
        self.assertTrue(torch.allclose(prob_or, log_or.exp()))
        a = log_nor.exp()+log_or.exp()
        ones = torch.ones_like(a)
        self.assertTrue(torch.allclose(a, ones))

class TestMultiBernoulliLogArithmetic(unittest.TestCase):
    def test_multi_bernoulli_equality(self):
        x = torch.tensor([-.5, .1, .9])
        y = torch.tensor([.5, .3, -.8])
        sig_x = torch.sigmoid(x)
        sig_y = torch.sigmoid(y)
        expected = sig_x*sig_y + (1-sig_x)*(1-sig_y)
        z0, z1 = multi_bernoulli_equality(x, y)
        self.assertTrue(torch.allclose(z1.exp(), expected))
        self.assertTrue(torch.allclose(z0.exp()+z1.exp(), torch.tensor(1.)))
    
    def test_multi_bernoulli_subset(self):
        x = torch.tensor([-.5, .1, .9])
        y = torch.tensor([.5, .3, -.8])
        sig_x = torch.sigmoid(x)
        sig_y = torch.sigmoid(y)
        expected = sig_x*sig_y + (1-sig_x)*(1-sig_y) + (1-sig_x)*sig_y
        z0, z1 = multi_bernoulli_subset(x, y)
        self.assertTrue(torch.allclose(z1.exp(), expected))
        self.assertTrue(torch.allclose(z0.exp()+z1.exp(), torch.tensor(1.)))
    
    def test_multi_bernoulli_activated_equality(self):
        x = torch.tensor([-.5, .1, .9])
        y = torch.tensor([.5, .3, -.8])
        a = torch.tensor([.5, .3, -.8])
        sig_x = torch.sigmoid(x)
        sig_y = torch.sigmoid(y)
        sig_a = torch.sigmoid(a)
        equal = sig_x*sig_y + (1-sig_x)*(1-sig_y)
        expected = sig_a + (1-sig_a)*equal
        z0, z1 = multi_bernoulli_activated_equality(x, y, a)
        self.assertTrue(torch.allclose(z1.exp(), expected))
        self.assertTrue(torch.allclose(z0.exp()+z1.exp(), torch.tensor(1.)))
    
    def test_multi_bernoulli_activated_subset(self):
        x = torch.tensor([-.5, .1, .9])
        y = torch.tensor([.5, .3, -.8])
        a = torch.tensor([.5, .3, -.8])
        sig_x = torch.sigmoid(x)
        sig_y = torch.sigmoid(y)
        sig_a = torch.sigmoid(a)
        subset = sig_x*sig_y + (1-sig_x)*(1-sig_y) + (1-sig_x)*sig_y
        expected = sig_a + (1-sig_a)*subset
        z0, z1 = multi_bernoulli_activated_subset(x, y, a)
        self.assertTrue(torch.allclose(z1.exp(), expected))
        self.assertTrue(torch.allclose(z0.exp()+z1.exp(), torch.tensor(1.)))


import unittest, torch
from radbm.metrics.hamming import hamming_pr_curve

class TestHammingPRCurve(unittest.TestCase):
    def test_hamming_pr_curve(self):
        queries = torch.tensor([
            [0,0,0,0],
            [0,0,1,1],
            [1,1,1,1],
        ])
        documents = torch.tensor([
            [0,0,0,1],
            [0,1,0,1],
            [1,0,1,1],
            [0,0,1,0],
        ])
        #hamming dists matrix
        #1, 2, 3, 1
        #1, 2, 1, 1
        #3, 2, 1, 3
        rel = [
            {0, 3}, #d is 1 1
            {1, 2}, #d is 2 1
            {0}, #d is 3
        ]
        #at 0 -> zero found -> p=nan, r=0
        #at 1 -> 6 found with 3 good out of 5 -> p=3/6=1/2, r=3/5
        #at 2 -> 9 found with 4 good out of 5 -> p=4/9, r=4/5
        #at 3 -> 12 found with 5 good out of 5 -> p=5/12, r=5/5=1
        #at 4 -> idem -> p=5/12, r=1
        expected_precisions = torch.tensor([float('nan'), 1/2, 4/9, 5/12, 5/12])
        expected_recalls = torch.tensor([0, 3/5, 4/5, 1, 1])
        precisions, recalls = hamming_pr_curve(documents, queries, rel)
        self.assertTrue(torch.isnan(precisions[0]))
        self.assertTrue(torch.allclose(precisions[1:], expected_precisions[1:]))
        self.assertTrue(torch.allclose(recalls, expected_recalls))
        
        #same thing but with return_valid_dists=True
        expected_dists = torch.tensor([1,2,3,4]) #not 0 because precision is undefined
        dists, precisions, recalls = hamming_pr_curve(documents, queries, rel, return_valid_dists=True)
        self.assertTrue(torch.allclose(dists, expected_dists))
        self.assertTrue(torch.allclose(precisions, expected_precisions[1:]))
        self.assertTrue(torch.allclose(recalls, expected_recalls[1:]))
        
        #error checks
        with self.assertRaises(ValueError):
            #code not same length
            hamming_pr_curve(documents[:,:-1], queries, rel, return_valid_dists=True)
        with self.assertRaises(ValueError):
            #not enough relevances
            hamming_pr_curve(documents, queries, rel[:-1], return_valid_dists=True)


import torch
from torch.nn.utils import clip_grad_value_
from torch.distributions import Categorical
from radbm.utils.torch import torch_soft_hamming
from radbm.search.elba import EfficientLearnableBinaryAccess

def categorical_entropy(cat):
    """
    -(cat*cat.log()).sum() without the annoying 0*inf
    
    Parameters
    ----------
    cat : torch.Tensor (ndim==1)
        The parameter of a Categorical distribution.
        
    Returns
    -------
    ent : torch.Tensor (a single float)
        The entropy of the Categorical distribution.
    """
    return Categorical(probs=cat).entropy()
    
def mi_categorical_bernoulli(pos_cat, neg_cat, p):
    """
    Compute the Multual Information between a categorical and a bernoulli.
    This use the fact that I(C, B) = H(C) - pH(C | B=1) - (1-p)H(C | B=0)
    with C = Cat(pi) and B = Ber(p).
    
    Parameters
    ----------
    pos_cat : torch.tensor (ndim=1, pos_cat.sum()=1)
        The parameters of C | B=1
    neg_cat : torch.tensor (ndim=1, neg_cat.sum()=1)
        The parameters of C | B=0
    p : float
        The parameters of B
        
    Returns
    -------
    I : torch.tensor (a single float)
            The Mutual Information I(C, B)
    """
    cat = p*pos_cat + (1-p)*neg_cat
    ent = categorical_entropy(cat)
    pos_ent = categorical_entropy(pos_cat)
    neg_ent = categorical_entropy(neg_cat)
    return ent - p*pos_ent - (1-p)*neg_ent

class TriangularKernel(torch.nn.Module):
    """
    Helper Module, compute the triangular kernel.
    """
    def __init__(self, centroids, widths=None):
        super().__init__()
        if widths is None:
            widths = torch.tensor(1, dtype=centroids.dtype)
        self.register_buffer('centroids', centroids)
        self.register_buffer('widths', widths)
        self.relu = torch.nn.ReLU()
        
    def forward(self, x):
        shape = x.shape
        x = x.view(*shape, 1)
        centroids = self.centroids.view(*len(shape)*[1], -1)
        return self.relu(1 - (centroids-x).abs()/self.widths)
    
class MIHash(EfficientLearnableBinaryAccess):
    """
    MIHash as in "MIHash: Online Hashing with Mutual Information"
    by Fatih Cakir, Kun He, Sarah Adel Bargal and Stan Sclaroff.

    Parameters
    ----------
    fq : torch.nn.Module
        The query Multi-Bernoulli encoder.
    fd : torch.nn.Module
        The document Multi-Bernoulli encoder.
    struct : BaseSDS subclass
        The structure used in ELBA.
    match_prob : float (in [0,1])
        The probability that there is a match given a random query
        and a random document.
    """
    def __init__(self, fq, fd, struct, nbits, match_prob, *args, **kwargs):
        super().__init__(fq, fd, struct, *args, **kwargs)
        self.match_prob = match_prob
        self.kernel = TriangularKernel(torch.arange(0,nbits+1))
    
    def step(self, q, d, match, l2_ratio=0):
        """
        Do a training step.
        
        Parameters
        ----------
        q : torch.Tensor
            A batch of queries.
        d : torch.Tensor
            A batch of documents.
        match : torch.Tensor (dtype=torch.bool)
            A matrix (2D tensor) with match[i,j] indicating if q[i] match with d[j]
            
        Returns
        -------
        loss : torch.Tensor (size 1)
            The loss (negative mutual information) of the current batch.
        """
        self.zero_grad()
        qsign = torch.tanh(self.fq(q))
        dsign = torch.tanh(self.fd(d))
        sh = torch_soft_hamming(qsign[:,None], dsign[None,:]) #shape = (#queries, #documents)
        bins = self.kernel(sh)
        pos_cat = bins[match].mean(dim=0)
        neg_cat = bins[~match].mean(dim=0)
        loss = -mi_categorical_bernoulli(pos_cat, neg_cat, self.match_prob)
        loss.backward()
        clip_grad_value_(self.parameters(), 5)
        self.optim.step()
        return loss


import unittest, torch
import numpy as np
from radbm.search.elba import Fbeta
from radbm.search.mbsds import HashingMultiBernoulliSDS

class TestFbeta(unittest.TestCase):
    def test_step(self):
        torch.manual_seed(0)
        model = Fbeta(
            torch.nn.Linear(32,16),
            torch.nn.Linear(32,16),
            HashingMultiBernoulliSDS(1,1),
            -np.log(32), # log(1/bs)
            match_dist=0,
        )
        x = torch.eye(32, dtype=torch.float)
        r = torch.eye(32, dtype=torch.bool)
        for i in range(500):
            model.step(x, x, r)
        model.batch_insert(x, range(32))
        out = model.batch_search(x)
        self.assertEqual(out, [{i} for i in range(32)])
        
    def test_raise(self):
        model = Fbeta(
            torch.nn.Linear(32,16),
            torch.nn.Linear(32,16),
            HashingMultiBernoulliSDS(1,1),
            -np.log(32), # log(1/bs)
            match_dist=0,
            nindex=3, #16%3 != 0
        )
        x = torch.eye(32, dtype=torch.float)
        r = torch.eye(32, dtype=torch.bool)
        with self.assertRaises(ValueError):
            model.step(x, x, r)


import unittest, colorsys, torch
from radbm.utils.torch import torch_hls_to_rgb, torch_rgb_to_hls

class TestTorchColor(unittest.TestCase):
    def test_torch_rgb_hls_convertions(self):
        torch.manual_seed(0xcafe)
        images = torch.rand(32,3,256,256)
        hls_images = torch_rgb_to_hls(images)
        some_pixels = [(1,2), (50,255), (42,42), (0,0)]
        for h, w in some_pixels:
            hls = torch.tensor(colorsys.rgb_to_hls(*images[0,:,h,w]))
            self.assertTrue(torch.allclose(hls_images[0,:,h,w], hls))
        rgb_images = torch_hls_to_rgb(hls_images)
        self.assertTrue(torch.allclose(images, rgb_images, atol=1e-6))


import itertools, torch
import numpy as np
from radbm.search.base import BaseSDS

def cast_numpy(data):
    if isinstance(data, torch.Tensor):
        return data.detach().cpu().numpy()
    return data

class HammingRadiusSDS(BaseSDS):
    """
    A Seach Data Structure for binary vectors, it searches within a Hamming radius.
    This uses hash tables to speed up the search.
    
    Parameters
    ----------
    nbits : int
        The number of bits the binary vectors will have.
    radius : int
        The Hamming radius to search inside. Should be small as the masks are precomputed.
        The memory taken is nbits * sum_{i=0}^radius comb(nbits, i).
    ntables : int
        The number of Hash tables to uses for the documents.
    """
    def __init__(self, nbits, radius, ntables=1):
        self.nbits = nbits
        self.radius = radius
        self.ntables = ntables
        comb = sum(
            (list(itertools.combinations(range(nbits),i)) for i in range(radius+1)),
            []
        )
        rows, cols = zip(*[(i,j) for i, c in enumerate(comb) for j in c])
        self.masks = np.zeros((len(comb), nbits), dtype=np.bool)
        self.masks[(rows, cols)] = True
        self.reset()
        
    def reset(self):
        """
        Empty the hash tables
        
        Returns
        -------
        self
        """
        self.tables = [dict() for _ in range(self.ntables)]
        return self
        
    def search(self, query):
        query = cast_numpy(query)
        indexes = set()
        for h in query^self.masks:
            th = tuple(bool(b) for b in h)
            for table in self.tables:
                if th in table:
                    indexes.update(table[th])
        return indexes
    
    def insert(self, document, index):
        document = cast_numpy(document)
        hashes = document^self.masks[:self.ntables]
        for h, table in zip(hashes, self.tables):
            th = tuple(bool(b) for b in h)
            if th in table: table[th].add(index)
            else: table[th] = {index}
        return self
    
    def itersearch(self, query, yield_empty=False):
        """
        Generator that search in the tables with a query in the form of a 
        Multi-Bernoulli distribution parametrized in log probabilities.
        This will search in each tables with each of the top (nlookups)
        outcomes. Everytime a set of indexes is found, this generator will
        yield it. If yield_empty is True, empty set will also be yield.

        Parameters
        ----------
        log_probs : numpy.ndarray (ndim == 1 or 2)
            If ndim==1: The Multi-Bernoulli distribution parametrized in log probabilities
            If ndim==2: len(log_probs) should be 2. The first element should be log
            probabilities that bits are zero and the second element should be the log
            probabilities that bits are one. This is for numerical stability
            (i.e. when probability are too close to 1)
        yield_empty : bool (optinal)
            Whether empty set should be yielded (default False)

        Yields
        ------
        indexes : set
            The newly found indexes
        """
        query = cast_numpy(query)
        yielded = set()
        for bits in query^self.masks:
            bits = tuple(bool(b) for b in bits)
            for table in self.tables:
                if bits in table:
                    indexes = table[bits]
                    new_indexes = indexes - yielded
                    yielded.update(new_indexes)
                    if yield_empty or new_indexes:
                        yield new_indexes
                elif yield_empty:
                    yield set()


import os
from radbm.utils import unique_list
from radbm.utils.gdrive import download_file

DATASETS_DIR = 'DATASETS_DIR'
MODELS_DIR = 'MODELS_DIR'
PACKAGE_VAR = 'PYTHONRADBM_DUMP'

def get_directories_list(path=None, data_type=None):
    """
    This function return the list of directory that is relevant for
    RADBM in order, it returns:

    path
    $DATASETS_DIR if data_type=='dataset'
    $MODELS_DIR if data_type=='model'
    $PYTHONRADBM_DUMP/<data_type>
    $HOME/.radbm/<data_type>
    .
    
    Parameters
    ----------
    path : str, optional
        The path to look first, this is helpful when a user specifies a
        path to file or relevant directory
    data_type : str, optional
        should be 'dataset' or 'model' this tell to lookup for
        environment variables specific to dataset or model
        
    Returns
    -------
    dirs : list of str
        The relevant directories in order
    """
    dirs = list()
    
    #path directory
    if path is not None:
        dirs.append(os.path.abspath(path))
    
    #data_type directories
    if data_type=='dataset':
        if DATASETS_DIR in os.environ:
            paths = os.environ[DATASETS_DIR].split(':')
            dirs.extend(unique_list(paths))
    elif data_type=='model':
        if MODELS_DIR in os.environ:
            paths = os.environ[MODELS_DIR].split(':')
            dirs.extend(unique_list(paths))
    elif data_type is not None:
        msg = 'data_type must be dataset or model, got {}'
        raise ValueError(msg.format(data_type))
    
    #package environment variable directory
    if PACKAGE_VAR in os.environ:
        path = os.environ[PACKAGE_VAR]
        if data_type is not None:
            path = os.path.join(path, data_type+'s')
        dirs.append(path)
        
    #package default direcotry
    if 'HOME' in os.environ:
        path = os.path.join(os.environ['HOME'], '.radbm')
        if data_type is not None:
            path = os.path.join(path, data_type+'s')
        dirs.append(path)
        
    #adding current direcory
    dirs.append(os.getcwd())
    
    return dirs

def expend_paths(paths, subpaths):
    """
    Parameters
    ----------
    dirs : list of str (directory path)
    subpaths : list of str (sub-directory path)
    """
    expended_paths = list()
    for path in paths:
        for subpath in subpaths:
            fullpath = os.path.join(path, subpath)
            expended_paths.append(fullpath)
    return expended_paths
    
def fetch_file(file, path=None, data_type=None, subdirs=None, download=True):
    """
    lookup on the machine for file otherwise download it.
    
    Parameters
    ----------
    file : str
        The name of the file to fetch
    path : str, optional (default=None)
        The principal path to look for the file. If the find is not
        found, this function will attemp to download to path to this file.
    data_type : str ('dataset' or 'model'), optional (default=None)
        Modifies the path to consider when looking for the file. If path is
        None and the file is not found, it will affect where to the file is
        downloaded, see get_directories_list for more information.
    subdirs : list of str, optional (default=None)
       Additional sub-directories to lookup
    download : bool
        A boolean to indicate if we want to download the file if not found
        on the machine
        
    Returns
    -------
    paths_list : list of str
        The list of path where to find the file
    """
    #Lookup
    dirs_list = get_directories_list(path, data_type=data_type)
    if subdirs is not None:
        dirs_list = expend_paths(dirs_list, subdirs)
    
    paths_list = expend_paths(dirs_list, [file])
    exists_paths_list = [path for path in paths_list if os.path.isfile(path)]
    
    if exists_paths_list:
        #file found
        return exists_paths_list
    elif download:
        #file not found -> download
        download_path = paths_list[0] #not empty, at least './<file>' in paths_list
        try:
            download_file(file, download_path)
        except KeyError as e:
            msg = '{} not found and is not registered for download'
            raise FileNotFoundError(msg.format(file)) from e
        return [download_path]
    else:
        msg = '{} not found and download disabled'
        raise FileNotFoundError(msg.format(file))


import unittest
import numpy as np
from scipy.stats import hypergeom

from radbm.utils.stats import *

np.seterr(all='raise')
def logsigmoid(x): return -np.log(1 + np.exp(-x))

class Test_least_k_subset_sum_generator(unittest.TestCase):
    def test_least_k_subset_sum_generator(self):
        rng = np.random.RandomState(0xcafe)
        values = rng.uniform(0, 100, (1000,))
        gen = least_k_subset_sum_generator(values, k=10**4)
        sums = [sum(values[list(subset)]) for subset in gen]
        self.assertTrue(np.all(0<=np.diff(sums))) #assert increasing order
        self.assertTrue(len(sums)==10**4)
        
        #assert generating all
        values = rng.uniform(0, 100, (10,))
        gen = least_k_subset_sum_generator(values)
        self.assertTrue(2**10, len(list(gen)))
    
def logsigmoid(x): return -np.log(1 + np.exp(-x))
class Test_greatest_k_multi_bernoulli_outcomes_generator(unittest.TestCase):
    def compute_outcomes_prob(self, generator, log_probs0, log_probs1):
        probs = list()
        for bits in generator:
            bits = np.array(bits)
            log_probs = np.where(bits, log_probs1, log_probs0)
            probs.append(np.exp(log_probs.sum()))
        return probs
            
    def test_multi_bernoulli_top_k_generato(self):
        rng = np.random.RandomState(0xcafe)
        logits = rng.uniform(-10, 10, (8,))
        log_probs0 = logsigmoid(logits)
        log_probs1 = logsigmoid(-logits)

        gen1 = greatest_k_multi_bernoulli_outcomes_generator(log_probs0, log_probs1, k=200)
        gen2 = greatest_k_multi_bernoulli_outcomes_generator(log_probs0, k=200)
        probs1 = self.compute_outcomes_prob(gen1, log_probs0, log_probs1)
        probs2 = self.compute_outcomes_prob(gen2, log_probs0, log_probs1)
        self.assertEqual(probs1, probs2)
        self.assertTrue(np.all(np.diff(probs1)<=0)) #assert descending order
        self.assertEqual(200, len(probs1)) #assert we generate k outcomes
        
        #assert that we generate all outcomes when k is None
        self.assertEqual(2**8, len(list(greatest_k_multi_bernoulli_outcomes_generator(log_probs0))))
        
        logits = rng.uniform(-100, 100, (8,))
        log_probs0 = logsigmoid(logits)
        log_probs1 = logsigmoid(-logits)
        gen1 = greatest_k_multi_bernoulli_outcomes_generator(log_probs0, log_probs1, k=200)
        gen2 = greatest_k_multi_bernoulli_outcomes_generator(log_probs0, k=200)
        next(gen1) #runs
        with self.assertRaises(FloatingPointError):
            #numerically unstable when log_probs1 is not given
            next(gen2)

class TestHypergeometric(unittest.TestCase):
    def test_hypergeometric(self):
        N = 20
        K = 3
        p = hypergeometric(N, K)
        scipy_p = np.array([hypergeom(N, K, n).pmf(range(0,K+1)) for n in range(N+1)])
        err = np.abs(p-scipy_p).max()
        self.assertTrue(err < 1e-10)

    def test_superdupergeometric_and_expectations(self):
        N = 10000
        K = 125
        sp = superdupergeometric(N, K)
        a = (sp*np.arange(N+1)[:,None]).sum(axis=0)
        b = b = superdupergeometric_expectations(N, K)
        err = np.abs(a[1:] - b[1:]).max()
        self.assertTrue(err < 1e-8)


from .utils import Test_unique_list, TestRamp
from .utils_fetch import Test_fetch_file
from .utils_torch import TorchCast
from .utils_torch_poisson_binomial import (
    TestLogPoissonBinomial,
    TestLogHammingBinomial,
)
from .utils_torch_multi_bernoulli_log_arithmetic import (
    TestLogAny,
    TestMultiBernoulliLogArithmetic,
)
    
from .utils_torch_color import TestTorchColor
from .utils_stats import (
    Test_least_k_subset_sum_generator,
    Test_greatest_k_multi_bernoulli_outcomes_generator,
    TestHypergeometric,
)
from .utils_time import TestChronometer
from .loaders_base import TestLoader, TestIRLoader
from .search_base import TestBaseSDS
from .search_mbsds import TestHashingMultiBernoulliSDS
from .search_gridsearch import TestGridSearch
from .search_radius import TestHammingRadiusSDS
from .search_elba_base import TestEfficientLearnableBinaryAccess
from .search_elba_fbeta import TestFbeta
from .search_elba_hbkl import TestHBKL
from .search_elba_mihash import TestMIHash
from .search_elba_hashnet import TestHashNet
from .metrics_oracle import TestOracleMetric
from .metrics_sswr import TestSSWR
from .metrics_hamming import TestHammingPRCurve


import unittest, torch
import numpy as np
from radbm.search.mbsds import HashingMultiBernoulliSDS

class TestHashingMultiBernoulliSDS(unittest.TestCase):
    def setUp(self):
        self.indexes = [0, 1, 2]
        self.docs_log_probs = np.log([
            [.3, .1, .8], #[0, 0, 1], [1, 0, 1], [0, 0, 0], ...
            [.7, .9, .2], #[1, 1, 0], [0, 1, 0], [1, 1, 1], ...
            [.3, .2, .1], #[0, 0, 0], [1, 0, 0], [0, 1, 0], ...
        ])
        self.qurs_log_probs = np.log([
            [.1, .4, .8], #[0, 0, 1], [0, 1, 1]
            [.1, .6, .2], #[0, 1, 0], [0, 0, 0]
        ])
        
    def test_numpy_cast(self):
        mbht = HashingMultiBernoulliSDS(3, 2)
        mbht.batch_insert(torch.tensor(self.docs_log_probs), self.indexes)
        
    def test_batch_insert(self):
        mbht = HashingMultiBernoulliSDS(3, 2)
        mbht.batch_insert(self.docs_log_probs, self.indexes)
        
    def test_bucket_stats(self):
        mbht = HashingMultiBernoulliSDS(3, 2)
        mbht.batch_insert(self.docs_log_probs, self.indexes)
        repr(mbht) #make sure it runs
        self.assertEqual(mbht.get_buckets_avg_size(), [1,1,1])
        self.assertEqual(mbht.get_buckets_max_size(), [1,1,1])
        
    def test_get_generator_error(self):
        ndim3 = np.zeros((1,2,3))
        mbht = HashingMultiBernoulliSDS(3, 2)
        with self.assertRaises(ValueError):
            mbht.insert(ndim3, 1)
        
    def test_itersearch(self):
        expected0 = [
            {0}, #with [0, 0, 1] -> {0}, {}, {},
                 #with [0, 1, 1] -> {}, {}, {},
        ]

        expected1 = [
            {1}, {2}, #with [0, 1, 0] -> {}, {1}, {2},
            {0}, #with [0, 0, 0] -> {2}, {}, {0}, (but 2 duplicate)
        ]
        
        mbht = HashingMultiBernoulliSDS(3, 2)
        mbht.batch_insert(self.docs_log_probs, self.indexes)
        result0 = list(mbht.itersearch(self.qurs_log_probs[0], nlookups=2))
        result1 = list(mbht.itersearch(self.qurs_log_probs[1], nlookups=2))
        self.assertEqual(result0, expected0)
        self.assertEqual(result1, expected1)
        
        #given log_probs1
        log_probs1 = np.log(1-np.exp(self.qurs_log_probs))
        log_probs = np.stack([self.qurs_log_probs, log_probs1], axis=1)
        result0 = list(mbht.itersearch(self.qurs_log_probs[0], nlookups=2))
        result1 = list(mbht.itersearch(self.qurs_log_probs[1], nlookups=2))
        self.assertEqual(result0, expected0)
        self.assertEqual(result1, expected1)
        
    def test_empty_itersearch(self):
        expected0 = [
            {0}, set(), set(), #with [0, 0, 1] -> {0}, {}, {},
            set(), set(), set(),  #with [0, 1, 1] -> {}, {}, {},
        ]

        expected1 = [
            set(), {1}, {2}, #with [0, 1, 0] -> {}, {1}, {2},
            set(), set(), {0}, #with [0, 0, 0] -> {2}, {}, {0}, (but 2 duplicate -> empty)
        ]
        
        mbht = HashingMultiBernoulliSDS(3, 2)
        mbht.batch_insert(self.docs_log_probs, self.indexes)
        result0 = list(mbht.itersearch(self.qurs_log_probs[0], nlookups=2, yield_empty=True))
        result1 = list(mbht.itersearch(self.qurs_log_probs[1], nlookups=2, yield_empty=True))
        self.assertEqual(result0, expected0)
        self.assertEqual(result1, expected1)
        
        #given log_probs1
        log_probs1 = np.log(1-np.exp(self.qurs_log_probs))
        log_probs = np.stack([self.qurs_log_probs, log_probs1], axis=1)
        result0 = list(mbht.itersearch(self.qurs_log_probs[0], nlookups=2, yield_empty=True))
        result1 = list(mbht.itersearch(self.qurs_log_probs[1], nlookups=2, yield_empty=True))
        self.assertEqual(result0, expected0)
        self.assertEqual(result1, expected1)
        
    def test_batch_search(self):
        expected = [
            {0}, #for query 0
            {0, 1, 2}, #for query 1
        ]
        mbht = HashingMultiBernoulliSDS(3, 2)
        mbht.batch_insert(self.docs_log_probs, self.indexes)
        result = mbht.batch_search(self.qurs_log_probs, nlookups=2)
        self.assertEqual(result, expected)
        
        #given log_probs1
        log_probs1 = np.log(1-np.exp(self.qurs_log_probs))
        log_probs = np.stack([self.qurs_log_probs, log_probs1], axis=1)
        result = list(mbht.batch_search(log_probs, nlookups=2))
        self.assertEqual(result, expected)
        
    def test_state(self):
        mbht = HashingMultiBernoulliSDS(3, 2)
        mbht.batch_insert(self.docs_log_probs, self.indexes)
        mbht_state = mbht.get_state()
        
        mbht_copy = HashingMultiBernoulliSDS(3, 2).set_state(mbht_state)
        self.assertEqual(mbht.tables, mbht_copy.tables)


import os, json, gdown
from pathlib import Path

_path = os.path.dirname(__file__)
_fileid_path = os.path.join(_path, 'fileid.json')
with open(_fileid_path, 'r') as f:
    _fileid = json.load(f)

    
def available_files():
    """
    Returns
    -------
    files : list of str
        The files available for download 
    """
    return list(_fileid.keys())
    
def download_file(file, path=None, verbose=False):
    """
    Download a file from Google Drive. This uses the gdown package.
    Warning, this function overwrite any existing file. 
    
    Parameters
    ----------
    file : str
        The name of the file to download. Use available_files() to 
        see which files are available for download.
    path : str, optional (default=file)
        The path where to download the file. If the path does not exists
        it will be created. This should contain the filename.
    verbose : bool, optional (default=False)
        If True output the download progress.
        
    Returns
    -------
    path : str
        The path where the file has been downloaded.
    """
    path = file if path is None else path
    directory = os.path.dirname(path)
    Path(directory).mkdir(parents=True, exist_ok=True)
    return gdown.download(_fileid[file], path, quiet=not verbose)


from radbm.utils.os import StateObj

class BaseSDS(StateObj):
    """
    Base Search Data Structure, need to be instantiate.
    Maintains an index for documents to be retrieved with query.
    When queried the appropriate index(es) will be returned (not
    the document(s))
    """
    def insert(self, document, index, *args, **kwargs):
        """
        Insert a single document in the data structure by saving the
        index. The document is not saved for scalability. This is
        the default implementation that uses batch_insert. See batch_insert
        documentation.
        
        Parameters
        ----------
            document : numpy.ndarray or torch.Tensor
                The document to by inserted (will not be saved)
            index : object
                A unique identifier for the document, some algorithms
                require index to be hashable.
            *args
                pass to batch_insert (see batch_insert for more details)
            **kwargs
                pass to batch_insert (see batch_insert for more details)
                
        """
        if type(self).batch_insert == BaseSDS.batch_insert:
            raise NotImplementedError('insert or batch_insert need to be overridden')
        self.batch_insert(document[None], [index], *args, **kwargs)
        return self
        
    def batch_insert(self, documents, indexes, *args, **kwargs):
        """
        Insert a multiple documents in the data structure by saving their
        index. The documents are not saved for scalability. This is
        the default implementation that uses insert. See insert
        documentation.
        
        Parameters
        ----------
            documents : numpy.ndarray or torch.Tensor
                The documents to by inserted (will not be saved). The first
                dimension is for the batch (i.e. documents[0] is a document)
            indexes : object
                A unique identifier for the document, some algorithms
                require index to be hashable.
            *args
                pass to insert (see insert for more details)
            **kwargs
                pass to insert (see insert for more details)
        """
        if type(self).insert == BaseSDS.insert:
            raise NotImplementedError('insert or batch_insert need to be overridden')
        for document, index in zip(documents, indexes):
            self.insert(document, index, *args, **kwargs)
        return self
            
    def search(self, query, *args, **kwargs):
        """
        Search in the data structure for the index of a documents. This is
        the default implementation that uses batch_search. See batch_search
        documentation.
        
        Parameters
        ----------
            query : numpy.ndarray or torch.Tensor
            *args
                pass to batch_search (see batch_search for more details)
            **kwargs
                pass to batch_search (see batch_search for more details)
            
        Returns
        -------
            indexes : set or list
                The indexes of the retrieved documents. If indexes
                is a list, it should indicate that the indexes are ordered.
        """
        if type(self).batch_search == BaseSDS.batch_search:
            raise NotImplementedError('search or batch_search need to be overridden')
        return self.batch_search(query[None], *args, **kwargs)[0]
    
    def batch_search(self, queries, *args, **kwargs):
        """
        Search in the data structure for the index of a documents. This is
        the default implementation that uses search. See search
        documentation.
        
        Parameters
        ----------
            query : numpy.ndarray or torch.Tensor
            *args
                pass to search (see search for more details)
            **kwargs
                pass to search (see search for more details)
            
        Returns
        -------
            indexes_list : list of (set or list)
                The indexes of the retrieved documents for each query.
                If indexes[i] is a list, it should indicate that the
                indexes are ordered.
        """
        if type(self).search == BaseSDS.search:
            raise NotImplementedError('search or batch_search need to be overridden')
        return [self.search(q, *args, **kwargs) for q in queries]
    
    def itersearch(self, query, *args, **kwargs):
        """
        Default generator, based on batch_itersearch, that yields indexes.
        
        Parameters
        ----------
            query : numpy.ndarray or torch.Tensor
            *args
                pass to batch_itersearch (see batch_itersearch for more details)
            **kwargs
                pass to batch_itersearch (see batch_itersearch for more details)
            
        Yields
        ------
            indexes
        """
        if type(self).batch_itersearch == BaseSDS.batch_itersearch:
            raise NotImplementedError('itersearch or batch_itersearch need to be overridden')
        yield from self.batch_itersearch(query[None], *args, **kwargs)[0]
        
    def batch_itersearch(self, queries, *args, **kwargs):
        """
        Default function that return a list of generator using itersearch.
        With ith generator corresponding to the ith query.
        
        Parameters
        ----------
            queries : numpy.ndarray or torch.Tensor
            *args
                pass to itersearch (see itersearch for more details)
            **kwargs
                pass to itersearch (see itersearch for more details)
            
        Returns
        ------
            itersearch_list : list of generator
                where the ith generetor yields indexes (set or list) that
                correspond to the ith  query.
        """
        if type(self).itersearch == BaseSDS.itersearch:
            raise NotImplementedError('itersearch or batch_itersearch need to be overridden')
        return [self.itersearch(query, *args, **kwargs) for query in queries]



import os, shutil, unittest
from pathlib import Path
from radbm.utils.gdrive import available_files
from radbm.utils.fetch import (
    fetch_file,
    DATASETS_DIR,
    MODELS_DIR,
    PACKAGE_VAR,
)

class Test_fetch_file(unittest.TestCase):
    def setUp(self):
        Path('./testdump').mkdir(exist_ok=True)
        with open('./testdump/existing_data', 'w'):
            pass
        
    def tearDown(self):
        shutil.rmtree('./testdump')
    
    def test_fetch_file(self):
        _environ = os.environ
        try:
            os.environ[DATASETS_DIR] = '/dataset_dir'
            os.environ[MODELS_DIR] = '/model_dir'
            os.environ[PACKAGE_VAR] = '/package_dir'
            os.environ['HOME'] = '/home_dir'

            path = './testdump/'
            results = fetch_file('existing_data', path, data_type='dataset')
            expected = [os.path.abspath('./testdump/existing_data')]
            self.assertEqual(results, expected)

            self.assertTrue('mnist.pkl.gz' in available_files())
            results = fetch_file('mnist.pkl.gz', path, data_type='model', subdirs=['Mnist'])
            expected = [os.path.abspath('./testdump/Mnist/mnist.pkl.gz')]
            self.assertEqual(results, expected)
            self.assertTrue(os.path.isfile('./testdump/Mnist/mnist.pkl.gz'))
            
            with self.assertRaises(ValueError):
                results = fetch_file('existing_data', path, data_type='wrong', download=True)

            with self.assertRaises(FileNotFoundError):
                results = fetch_file('missing_data', path, data_type='dataset', download=True)

            with self.assertRaises(FileNotFoundError):
                results = fetch_file('missing_data', path, data_type='dataset', download=False)
        finally:
            os.environ.clear()
            os.environ.update(_environ)


import unittest, torch
import numpy as np
from radbm.search.elba import EfficientLearnableBinaryAccess
from radbm.search.mbsds import HashingMultiBernoulliSDS

class TestEfficientLearnableBinaryAccess(unittest.TestCase):
    def test_elba(self):
        f = torch.nn.Linear(784,128)
        elba = EfficientLearnableBinaryAccess(
            f, f, HashingMultiBernoulliSDS(1,1))
        data = np.random.RandomState(0xcafe).normal(0,1,(32, 784))
        data = torch.tensor(data, dtype=torch.float32)
        elba.batch_insert(data, range(32))
        #using the same function for fq and fd
        index = elba.batch_search(data)
        self.assertEqual(index, [{i} for i in range(32)])
        index = [next(g) for g in elba.batch_itersearch(data)]
        self.assertEqual(index, [{i} for i in range(32)])
        
    def test_get_and_set_state(self):
        f = torch.nn.Linear(784,128)
        elba = EfficientLearnableBinaryAccess(
            f, f, HashingMultiBernoulliSDS(1,1))
        elba.set_state(elba.get_state())


import numpy as np
import gzip, pickle, torch
from radbm.loaders.base import Loader, IRLoader
from radbm.utils.fetch import get_directories_list, fetch_file

def mnist_loader(path=None, download=True):
    file_paths = fetch_file('mnist.pkl.gz', path, data_type='dataset', subdirs=['Mnist', 'mnist'], download=download)
    if not file_paths:
        raise FileNotFoundError('could not locate mnist.pkl.gz')
    with gzip.open(file_paths[0], 'rb') as f:
        train_xy, valid_xy, test_xy = pickle.load(f, encoding='latin1')
    return train_xy, valid_xy, test_xy
    
class Mnist(Loader):
    def __init__(self, path=None, download=True, which='train', backend='numpy', device='cpu', rng=np.random):
        super().__init__(which=which, backend=backend, device=device, rng=rng)
        train_xy, valid_xy, test_xy = mnist_loader(path, download=download)
        for which, (x, y) in [('train', train_xy), ('valid', valid_xy), ('test', train_xy)]:
            self.register_switch('{}_x'.format(which), x)
            self.register_switch('{}_y'.format(which), y)
            x = getattr(self, '{}_x'.format(which)) #retrieve the TorchNumpy
            y = getattr(self, '{}_y'.format(which)) #retrieve the TorchNumpy
            setattr(self, '{}_group'.format(which), {'x':x, 'y':y})
        getattr(self, self.which)()
        
    def batch(self, size=None, index=None, replace=True):
        if index is not None or not replace:
            raise NotImplementedError('not implemented yet')
        ids = self.rng.randint(0, len(self.y.data), size)
        return self.x.data[ids], self.y.data[ids]
    
class ClassMnist(IRLoader):
    def __init__(self, sigma=0, path=None, download=True, mode='class_relation', which='train', backend='numpy', device='cpu', rng=np.random):
        super().__init__(mode=mode, which=which, backend=backend, device=device, rng=rng)
        if mode != 'relational_matrix':
            raise NotImplementedError('mode={} not implemented yet'.format(mode))
        self.sigma = sigma
        train_xy, valid_xy, test_xy = mnist_loader(path, download=download)
        for which, (x, y) in [('train', train_xy), ('valid', valid_xy), ('test', train_xy)]:
            self.register_switch('{}_x'.format(which), x[y.argsort()])
            jump = np.arange(0,len(x),len(x)//10) #jump always on np since it is for indexing
            x = getattr(self, '{}_x'.format(which)) #retrieve the TorchNumpy
            setattr(self, '{}_group'.format(which), {'x':x, 'jump':jump})
        self.register_switch('eye', np.eye(10, dtype=np.bool))
        getattr(self, self.which)()
        
    def get_available_modes(self):
        return {
            'class_relation',
            'relational_triplets',
            'relational_matrix'
        }
        
    def get_relation_prob(self):
        return 1/10
    
    def get_relation_log_prob(self):
        return -np.log(10)
        
    def build_any(self, size=None, index=None, replace=True):
        if size is not None or replace:
            raise NotImplementedError('not implemented yet')
        x = self.x.data if index is None else self.x.data[index]
        return x + self.rng.normal(0, self.sigma, x.shape)
    
    def build_queries(self, size=None, index=None, replace=True):
        return self.build_any(size, index, replace) #because the tasks is symmetric
    
    def build_documents(self, size=None, index=None, replace=True):
        return self.build_any(size, index, replace) #because the tasks is symmetric
    
    def batch(self, size=None, index=None, replace=True):
        if size != 10: raise NotImplementedError('batch size != 10 not supported')
        x = self.x.data
        ids1 = self.rng.rng.randint(0, len(x)//10, size) + self.jump
        ids2 = self.rng.rng.randint(0, len(x)//10, size) + self.jump
        x1, x2 = x[ids1], x[ids2]
        x1 = x1 + self.rng.normal(0, self.sigma, x1.shape)
        x2 = x2 + self.rng.normal(0, self.sigma, x2.shape)
        return x1, x2, self.eye.data
    
class NoisyMnist(IRLoader):
    def __init__(self, sigma=0.2, path=None, download=True, mode='relational_matrix', which='train', backend='numpy', device='cpu', rng=np.random):
        super().__init__(mode=mode, which=which, backend=backend, device=device, rng=rng)
        if mode != 'relational_matrix':
            raise NotImplementedError('mode={} not implemented yet'.format(mode))
        self.sigma = sigma
        train_xy, valid_xy, test_xy = mnist_loader(path, download=download)
        for which, (x, y) in [('train', train_xy), ('valid', valid_xy), ('test', train_xy)]:
            self.register_switch('{}_x'.format(which), x)
            x = getattr(self, '{}_x'.format(which)) #retrieve the TorchNumpy
            setattr(self, '{}_group'.format(which), {'x':x})
        getattr(self, self.which)()
    
    def get_available_modes(self):
        return {
            'relational_pairs', #(x,k,r)
            'relational_triplets', #(x, pos_x, neg_x)
            'relational_matrix', #(xs, ks, R)
        }
    
    def get_relation_prob(self):
        return 1/len(self.x.data)
    
    def get_relation_log_prob(self):
        return -np.log(len(self.x.data))
    
    def iter_any(self, batch_size, seed, maximum):
        which = self.which
        N = min(maximum, len(self.x.data))
        nbatch = int(np.ceil(N//batch_size))
        rng = np.random.RandomState(seed)
        for start in [i*batch_size for i in range(nbatch)]:
            if self.which != which:
                msg = 'self.which changed while iterating, should be "{}" but is now "{}".'
                raise RuntimeError(msg.format(which, self.which))
            end = min(start+batch_size, N)
            data = self.x.data[start:end].reshape(-1, 28, 28)
            data = data + self.dynamic_cast(rng.normal(0, self.sigma, data.shape))
            yield data, list(range(start, end))
            
    def iter_queries(self, batch_size, maximum=np.inf):
        for data, ids in self.iter_any(batch_size, seed=0xcafe, maximum=maximum):
            yield data, [{i} for i in ids]
    
    def iter_documents(self, batch_size, maximum=np.inf):
        yield from self.iter_any(batch_size, seed=0xbeef, maximum=maximum)
    
    def batch(self, size=None, index=None, replace=True):
        if index is not None or not replace:
            raise ValueError('not implemented yet')
        if size is None: size = len(self.x.data)
        ids = self.rng.rng.randint(0, len(self.x.data), size)
        x = self.x.data[ids]
        q = x + self.rng.normal(0, self.sigma, x.shape)
        d = x + self.rng.normal(0, self.sigma, x.shape)
        r = self.dynamic_cast(np.eye(size, dtype=np.bool))
        return q, d, r
    
def _reverse_index(index):
    r = dict()
    for n, i in enumerate(index):
        for j in i.flatten():
            if j in r: r[j].add(n)
            else: r[j] = {n}
    return [r[i] for i in range(10000)]

class MnistMosaic(IRLoader):
    def __init__(self, sigma=0.2, path=None, download=True, mode='relational_matrix', which='train', backend='numpy', device='cpu', rng=np.random):
        super().__init__(mode=mode, which=which, backend=backend, device=device, rng=rng)
        if mode != 'relational_matrix':
            raise NotImplementedError('mode={} not implemented yet'.format(mode))
        self.sigma = sigma
        dbrng = np.random.RandomState(0xcafe) #used only for indexes generation
        train_xy, valid_xy, test_xy = mnist_loader(path, download=download)
        for which, (x, y) in [('train', train_xy), ('valid', valid_xy), ('test', train_xy)]:
            self.register_switch('{}_x'.format(which), x)
            x = getattr(self, '{}_x'.format(which)) #retrieve the TorchNumpy
            documents_indexes = dbrng.randint(0, 10000, (100000, 2, 2))
            axis1, axis2 = np.random.randint(0,2,(2, 10000))
            documents_indexes[range(10000), axis1, axis2] = range(10000) #so the n query is relevant the the n document
            relevants = _reverse_index(documents_indexes) #might crash with "bad" generator seed
            setattr(self, '{}_group'.format(which), {
                'x' :x,
                'documents_indexes': documents_indexes,
                'relevants': relevants,
            })
        getattr(self, self.which)()
    
    def get_available_modes(self):
        return {
            'relational_pairs', #(x,k,r)
            'relational_triplets', #(x, pos_x, neg_x)
            'relational_matrix', #(xs, ks, R)
        }
    
    def get_relation_prob(self):
        return 1/len(self.x.data)
    
    def get_relation_log_prob(self):
        return -np.log(len(self.x.data))
                    
    def dynamic_backend_mosaic(self, ids):
        x = self.x.data[ids.flatten()]
        if self.backend == 'torch':
            x = x.view(-1, 2, 2, 28, 28).permute(0,1,3,2,4).reshape(-1, 56, 56)
        elif self.backend == 'numpy':
            x = x.reshape(-1, 2, 2, 28, 28).transpose(0,1,3,2,4).reshape(-1, 56, 56)
        return x
    
    def iter_queries(self, batch_size, maximum=np.inf):
        which = self.which
        N = min(maximum, len(self.x.data))
        nbatch = int(np.ceil(N//batch_size))
        rng = np.random.RandomState(0xcafe)
        for start in [i*batch_size for i in range(nbatch)]:
            if self.which != which:
                msg = 'self.which changed while iterating, should be "{}" but is now "{}".'
                raise RuntimeError(msg.format(which, self.which))
            end = min(start+batch_size, N)
            data = self.x.data[start:end].reshape(-1, 28, 28)
            data = data + self.dynamic_cast(rng.normal(0, self.sigma, data.shape))
            relevants = self.relevants[start:end]
            yield data, relevants
    
    def iter_documents(self, batch_size, maximum=np.inf):
        which = self.which
        N = min(maximum, len(self.documents_indexes))
        nbatch = int(np.ceil(N//batch_size))
        rng = np.random.RandomState(0xbeef)
        for start in [i*batch_size for i in range(nbatch)]:
            if self.which != which:
                msg = 'self.which changed while iterating, should be "{}" but is now "{}".'
                raise RuntimeError(msg.format(which, self.which))
            end = min(start+batch_size, N)
            indexes = self.documents_indexes[start:end]
            data = self.dynamic_backend_mosaic(indexes)
            data = data + self.dynamic_cast(rng.normal(0, self.sigma, data.shape))
            ids = list(range(start, end))
            yield data, ids
    
    def batch(self, size=None, index=None, replace=True):
        if index is not None or not replace:
            raise ValueError('not implemented yet')
        if size is None: size = len(self.x.data)
            
        d_ids = self.rng.rng.randint(0, len(self.x.data), (size,2,2))
        d = self.dynamic_backend_mosaic(d_ids)
        d = d + self.rng.normal(0, self.sigma, d.shape)
        
        axis1, axis2 = np.random.randint(0,2,(2, size))
        q_ids = d_ids[range(size), axis1, axis2]
        q = self.x.data[q_ids].reshape(size, 28, 28)
        q = q + self.rng.normal(0, self.sigma, q.shape)
        
        r = (q_ids[:,None,None,None] == d_ids[None]).any(axis=(2,3))
        r = self.dynamic_cast(r)
        return q, d, r


import os, torch

#not safe yet
def safe_save(obj, path, pickle_protocol=torch.serialization.DEFAULT_PROTOCOL):
    """
    Safely saves obj in path by creating .tmp file
    before ovewriting an existing file

    Parameters
    ----------
    obj : object
        the object to be saved
    path : str
        the path where to save obj
    pickle_protocol : int (optional)
        The pickle protocol to uses. By default, the torch default_protocol
        is used.
    
    Returns
    -------
    obj : object
        the same obj received as input

    """
    with open(path, 'wb') as f:
        torch.save(obj, f, pickle_protocol=pickle_protocol)
    return obj
    
#not safe yet
def safe_load(path, map_location=None):
    """
    Safely load an object if .tmp is present 
    it will be prefered unless it is corrupted
    (i.e. pickle.load fails)

    Parameters
    ----------
    path : str
        the path where to object is saved
    
    Returns
    -------
    obj : object
        the object loaded

    """
    with open(path, 'rb') as f:
        obj = torch.load(f, map_location=map_location)
    return obj

class StateObj(object):
    def get_state(self):
        raise NotImplementedError()
    
    def set_state(self, state):
        raise NotImplementedError()
    
    def save(self, path, if_exists='overwrite', *args, **kwargs):
        #TODO replace *args **kwargs with the right names
        if not os.path.exists(path) or if_exists=='overwrite':
            state = self.get_state()
            safe_save(state, path, *args, **kwargs)
        elif if_exists=='raise':
            msg = '{} exists and if_exists is "raise"'
            raise FileExistsError(msg.format(path))
        return self
    
    def load(self, path, if_inexistent='raise', *args, **kwargs):
        #TODO replace *args **kwargs with the right names
        if os.path.exists(path):
            state = safe_load(path, *args, **kwargs)
            self.set_state(state)
        elif if_inexistent=='raise':
            msg = '{} does not exists and if_inexistent is "raise"'
            raise FileNotFoundError(msg.format(path))
        return self



from time import time

class Chronometer(object):
    """
    A chronometer found time codes.

    Methods
    -------
    reset()
        Resets the chronometer.
    start()
        Starts the chronometer.
    stop()
        Stops the chronometer.
    time() : float
        return the number of second on the chronometer.

    """
    def __init__(self):
        self.reset()
        
    def reset(self):
        self.running = False
        self.counter = 0
        return self
        
    def start(self):
        if self.running:
            raise ValueError('cannot start if running')
        self.running = True
        self.t0 = time()
        return self
        
    def stop(self):
        if not self.running:
            raise ValueError('cannot stop if not running')
        self.counter += time() - self.t0
        self.running = False
        return self
    
    def time(self):
        if self.running:
            return self.counter + time() - self.t0
        return self.counter
    
    def __repr__(self):
        state = 'start' if self.running else 'stop'
        return '{:.4f} ({})'.format(self.time(), state)